from Configurables import DaVinci, GaudiSequencer

from helpers import tuple_maker

name = 'Bu2JpsiKplus'
line = 'BetaSBu2JpsiKDetachedLine'
is_mc = DaVinci().Simulation
tuple_seq = tuple_maker.tuple_maker(
    name,
    decay='[B+ -> ^(J/psi(1S) -> ^mu+ ^mu-) ^K+]CC',
    branches={
        'B'       : '[B+ ->  (J/psi(1S) ->  mu+  mu-)  K+]CC',
        'Jpsi'    : '[B+ -> ^(J/psi(1S) ->  mu+  mu-)  K+]CC',
        'muplus'  : '[B+ ->  (J/psi(1S) -> ^mu+  mu-)  K+]CC',
        'muminus' : '[B+ ->  (J/psi(1S) ->  mu+ ^mu-)  K+]CC',
        'hplus'   : '[B+ ->  (J/psi(1S) ->  mu+  mu-) ^K+]CC'
    },
    stripping_line=line,
    is_mc=is_mc,
    input_type=DaVinci().InputType
)
dtt = tuple_seq.Members[-1]

# Alternate mass hypotheses
fitter_configs = [
    # B -> KplusPiMuMu
    ('B2PiJpsi', {
        'Beauty -> Meson ^K+': 'pi+',
        'Beauty -> Meson ^K-': 'pi-'
    }),
    # Lb -> pKMuMu (Kplus)
    ('B2pJpsi', {
        'Beauty -> Meson ^K+': 'p+',
        'Beauty -> Meson ^K-': 'p~-'
    })
]
for fitter_name, substitutions in fitter_configs:
    fitter = dtt.B.addTupleTool('TupleToolDecayTreeFitter/{}'.format(fitter_name))
    fitter.Verbose = True
    fitter.Substitutions = substitutions
    fitter.daughtersToConstrain = ['J/psi(1S)']
    fitter.constrainToOriginVertex = True

docas = dtt.B.addTupleTool('TupleToolDOCA')
doca_name, location1, location2 = zip(
    *[[
          'muplus_muminus',
          '[B+ ->  (J/psi(1S) ->  ^mu+   mu-)   K+]CC',
          '[B+ ->  (J/psi(1S) ->   mu+  ^mu-)   K+]CC'
      ],
      [
          'hplus_muplus',
          '[B+ ->  (J/psi(1S) ->   mu+   mu-)  ^K+]CC',
          '[B+ ->  (J/psi(1S) ->  ^mu+   mu-)   K+]CC'
      ],
      [
          'hplus_muminus',
          '[B+ ->  (J/psi(1S) ->   mu+   mu-)  ^K+]CC',
          '[B+ ->  (J/psi(1S) ->   mu+  ^mu-)   K+]CC'
      ]])
docas.Name = list(doca_name)
docas.Location1 = list(location1)
docas.Location2 = list(location2)

mc_tuples = []
if is_mc:
    mc_tuples.append(tuple_maker.mc_tuple_maker(
        name,
        '[[B+]cc ==> ^K+ ^(J/psi(1S) ==> ^mu+ ^mu-)]CC',
        {
            'B'       : '[[B+]cc ==>  K+  (J/psi(1S) ==>  mu+  mu-)]CC',
            'Jpsi'    : '[[B+]cc ==>  K+ ^(J/psi(1S) ==>  mu+  mu-)]CC',
            'muplus'  : '[[B+]cc ==>  K+  (J/psi(1S) ==> ^mu+  mu-)]CC',
            'muminus' : '[[B+]cc ==>  K+  (J/psi(1S) ==>  mu+ ^mu-)]CC',
            'hplus'   : '[[B+]cc ==> ^K+  (J/psi(1S) ==>  mu+  mu-)]CC'
        }
    ))

seq = tuple_maker.tuple_sequence()
seq.Members += [tuple_seq] + mc_tuples



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import pytest
from click.testing import CliRunner

from LbAPLocal.cli import legacy_main as main


def test_list(data_pkg_repo):
    runner = CliRunner()
    result = runner.invoke(main, ["list"])
    assert result.exit_code == 0, result.stdout
    assert "cbtesting" in result.stdout
    assert "BsToJpsiPhi" in result.stdout
    assert "RDs" in result.stdout


@pytest.mark.parametrize(
    "prod_name,example,num_expected",
    [
        ("cbtesting", "MC_2016_MagDown_Lb2PsiL_mm_tuple", 168),
        ("BsToJpsiPhi", "MC_2018_MagUp_InclJpsi_sim09g_s34_dst", 16),
    ],
)
def test_list_with_production(data_pkg_repo, prod_name, example, num_expected):
    runner = CliRunner()
    result = runner.invoke(main, ["list", prod_name])
    assert result.exit_code == 0, result.stdout
    assert example in result.stdout
    assert result.stdout.count("\n*") == num_expected


@pytest.mark.parametrize("prod_name", ["cbtesting", "BsToJpsiPhi"])
def test_render(data_pkg_repo, prod_name):
    runner = CliRunner()
    result = runner.invoke(main, ["render", prod_name])
    assert result.exit_code == 0, result.stdout


@pytest.mark.parametrize("prod_name", ["RDs"])
def test_render_bad(data_pkg_repo, prod_name):
    runner = CliRunner()
    result = runner.invoke(main, ["render", prod_name])
    assert result.exit_code == 1
    assert "Rendered YAML has errors!" in result.stdout


@pytest.mark.parametrize("prod_name", ["cbtesting", "BsToJpsiPhi"])
def test_validate(data_pkg_repo, prod_name):
    runner = CliRunner()
    result = runner.invoke(main, ["validate", prod_name])
    assert result.exit_code == 0, result.stdout


@pytest.mark.parametrize("prod_name,expected_message", [("RDs", "output:")])
def test_validate_bad(data_pkg_repo, prod_name, expected_message):
    runner = CliRunner()
    result = runner.invoke(main, ["validate", prod_name])
    assert result.exit_code == 1
    assert "Error parsing YAML!" in result.stdout



###############################################################################
# (c) Copyright 2020-2024 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Core logic for creating minimal reproducers from failed grid jobs.

Decomposes the slicing pipeline into independently testable functions:
  1. Find the failed step in a downloaded job directory
  2. Parse Gaudi logs for event information
  3. Create modified prodConf files for reproducing failures
  4. Run the reproducer and analyze output
  5. Extract single events and package minimal reproducers
"""

import json
import re
import shutil
import subprocess
import threading
import xml.etree.ElementTree as ET
from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path

import yaml


@dataclass
class InputFileStatus:
    """Status of an input file from the summary XML.

    Attributes:
        lfn: The LFN of the input file (e.g. "LFN:/lhcb/...").
        status: "full" if fully processed, "part" if partially processed.
        events_read: Number of events read from this file.
    """

    lfn: str
    status: str
    events_read: int | None = None


@dataclass
class FailedStep:
    """Information about a failed step in a job."""

    step_id: str
    xml_file: Path
    log_file: Path
    prodconf_file: Path | None = None
    input_file_statuses: list[InputFileStatus] | None = None


@dataclass
class EventInfo:
    """Information about the last event processed before failure.

    Attributes:
        event_number: The overall event number (across all input files).
        slot: The processing slot number.
        can_slice_single: True if new-style HLTControlFlowMgr SUCCESS messages
            were found, allowing slicing down to a single event. False if only
            old-style RootIOAlg SUCCESS messages were found (coarser slicing).
        event_in_file: The in-file event number (only set for old-style messages).
        lfn: The LFN of the file being read (only set for old-style messages).
        suspect_lfns: LFNs most likely to contain the error.
        suspect_event_in_file: Best guess of in-file event number, or None
            if we can't narrow it down to one event in one file.
        failed_event_overall: Event number from HLTControlFlowMgr FATAL.
    """

    event_number: int
    slot: int
    can_slice_single: bool = True
    event_in_file: int | None = None
    lfn: str | None = None
    unprocessed_lfns: list[str] = field(default_factory=list)
    suspect_lfns: list[str] = field(default_factory=list)
    suspect_event_in_file: int | None = None
    failed_event_overall: int | None = None


@dataclass
class ReproducerOutput:
    """Parsed output from running a reproducer.

    Attributes:
        last_event_in_file: Last in-file event number from RootIOAlg (old style).
        last_event_lfn: LFN of the last-read file from RootIOAlg (old style).
        last_event_overall: Last overall event count from RootIOAlg (old style).
        last_processing_event: Last event from HLTControlFlowMgr SUCCESS (new style).
        failed_event_overall: Event number from HLTControlFlowMgr FATAL (bonus).
        exact_problem_event: Calculated exact in-file event that caused failure.
        can_slice_single: Whether new-style messages allow single-event precision.
        returncode: Exit code from the reproducer run.
    """

    last_event_in_file: int | None = None
    last_event_lfn: str | None = None
    last_event_overall: int | None = None
    last_processing_event: int | None = None
    failed_event_overall: int | None = None
    exact_problem_event: int | None = None
    can_slice_single: bool = False
    returncode: int = 0


CVMFS_DBASE = Path("/cvmfs/lhcb.cern.ch/lib/lhcb/DBASE")


@dataclass
class DataPackage:
    """A data package directory from CVMFS to include in the reproducer."""

    name: str
    version: str
    cvmfs_path: Path


def find_failed_step(job_dir: Path) -> FailedStep:
    """Scan summary*.xml files for a step with <success>false</success>.

    Args:
        job_dir: Path to the downloaded job directory.

    Returns:
        FailedStep with step_id, xml_file, log_file, and optionally prodconf_file.

    Raises:
        FileNotFoundError: If no XML summary files are found.
        ValueError: If no failed steps are found.
    """
    xml_files = sorted(job_dir.glob("summary*.xml"))
    if not xml_files:
        raise FileNotFoundError(f"No XML summary files found in {job_dir}")

    for xml_file in xml_files:
        try:
            tree = ET.parse(xml_file)
            root = tree.getroot()
            success_elem = root.find(".//success")

            if (
                success_elem is not None
                and success_elem.text
                and success_elem.text.strip().lower() == "false"
            ):
                step_id = xml_file.stem.replace("summary", "")
                log_file = job_dir / f"{step_id}.log"

                # Look for corresponding prodConf JSON
                prodconf_file = job_dir / f"prodConf_{step_id}.json"
                if not prodconf_file.exists():
                    # Try alternative naming patterns
                    prodconf_candidates = list(
                        job_dir.glob(f"prodConf*{step_id}*.json")
                    )
                    prodconf_file = (
                        prodconf_candidates[0] if prodconf_candidates else None
                    )

                input_file_statuses = parse_input_file_statuses(root)

                return FailedStep(
                    step_id=step_id,
                    xml_file=xml_file,
                    log_file=log_file,
                    prodconf_file=prodconf_file,
                    input_file_statuses=input_file_statuses or None,
                )
        except ET.ParseError:
            continue

    raise ValueError("No failed steps found in XML summaries")


def parse_input_file_statuses(root: ET.Element) -> list[InputFileStatus]:
    """Parse input file statuses from a summary XML root element.

    The summary XML has a structure like::

        <input>
          <file name="LFN:..." status="full">947185</file>
          <file name="LFN:..." status="part">149417</file>
        </input>

    Files with ``status="full"`` were fully processed.  Files with
    ``status="part"`` were partially processed (the error likely occurred
    here).  Files listed in the prodConf but absent from the XML were
    never opened.

    Args:
        root: The parsed XML root element.

    Returns:
        List of InputFileStatus entries.  Empty if the XML has no
        ``<input>`` section or no ``<file>`` elements with status
        attributes (e.g. after a segfault).
    """
    input_elem = root.find(".//input")
    if input_elem is None:
        return []

    statuses = []
    for file_elem in input_elem.findall("file"):
        name = file_elem.get("name", "")
        status = file_elem.get("status", "")
        if not name or not status:
            continue

        events_read = None
        if file_elem.text and file_elem.text.strip().isdigit():
            events_read = int(file_elem.text.strip())

        statuses.append(
            InputFileStatus(
                lfn=name,
                status=status,
                events_read=events_read,
            )
        )

    return statuses


def find_last_event(
    log_file: Path,
    input_file_statuses: list[InputFileStatus] | None = None,
) -> EventInfo:
    """Parse Gaudi log for event progress messages.

    Tries two patterns in order of preference:

    1. **New style** (can slice to single event)::

        HLTControlFlowMgr SUCCESS Processing event N in slot M

       For new style, also tracks LFNs that were opened but not fully processed:
       - Finds all LFNs from: RootIOAlg INFO DATAFILE='LFN:...'
       - Finds all LFNs from: RootIOHandler INFO Fully processed input file LFN:...
       - The difference is stored in EventInfo.unprocessed_lfns for testing

    2. **Old style** (can only slice to ~100-event window)::

        RootIOAlg SUCCESS Reading Event record N within LFN:..., M overall

    **Run 2 style**: If Run 2 EventSelector messages are detected, raises NotImplementedError
    as precise slicing is not yet implemented for Run 2 jobs::

        EventSelector SUCCESS Reading Event record M. Record number within stream 1: N

    Also checks for the bonus FATAL pattern::

        HLTControlFlowMgr FATAL Event N on slot M failed!

    Args:
        log_file: Path to the step log file.

    Returns:
        EventInfo with event_number, slot, and can_slice_single flag.
        For new-style logs, also includes unprocessed_lfns list.

    Raises:
        FileNotFoundError: If the log file doesn't exist.
        ValueError: If no event processing messages are found.
        NotImplementedError: If Run 2 style messages are detected.
    """
    if not log_file.exists():
        raise FileNotFoundError(f"Log file not found: {log_file}")

    new_style = re.compile(
        r"HLTControlFlowMgr\s+SUCCESS\s+Processing event\s+(\d+)\s+in slot\s+(\d+)"
    )  # DV v66r9
    old_style = re.compile(
        r"RootIOAlg\s+SUCCESS\s+Reading Event record\s+(\d+)\s+within\s+(LFN:[^,]+),\s+(\d+)\s+overall"
    )
    run2_style = re.compile(
        r"EventSelector\s+SUCCESS\s+Reading Event record\s+(\d+)\.\s+Record number within stream (\d+):\s+(\d+)"
    )
    run2_lfn = re.compile(
        r"EventSelector\s+INFO\s+Stream:EventSelector\.DataStreamTool_(\d+)\s+Def:DATAFILE='(LFN:[^']+)'"
    )
    # Track LFNs opened vs fully processed
    rootioalg_datafile = re.compile(r"RootIOAlg\s+INFO\s+DATAFILE='(LFN:[^']+)'")
    rootiohandler_processed = re.compile(
        r"RootIOHandler\s+INFO\s+Fully processed input file\s+(LFN:\S+)"
    )
    fatal_pattern = re.compile(
        r"HLTControlFlowMgr\s+FATAL\s+Event\s+(\d+)\s+on\s+slot\s+(\d+)\s+failed"
    )

    last_new = None
    last_old = None
    last_run2 = None
    failed_event: int | None = None
    # Map stream number to LFN for Run 2 style
    stream_to_lfn: dict[int, str] = {}
    # Track opened and processed LFNs
    opened_lfns: set[str] = set()
    processed_lfns: set[str] = set()

    with open(log_file) as f:
        for line in f:
            m = new_style.search(line)
            if m:
                last_new = EventInfo(
                    event_number=int(m.group(1)),
                    slot=int(m.group(2)),
                    can_slice_single=True,
                )

            m = old_style.search(line)
            if m:
                last_old = EventInfo(
                    event_number=int(m.group(3)),  # overall count
                    slot=0,
                    can_slice_single=False,
                    event_in_file=int(m.group(1)),
                    lfn=m.group(2),
                )

            m = run2_lfn.search(line)
            if m:
                stream_num = int(m.group(1))
                lfn = m.group(2)
                stream_to_lfn[stream_num] = lfn

            m = run2_style.search(line)
            if m:
                stream_num = int(m.group(2))
                last_run2 = EventInfo(
                    event_number=int(m.group(1)),  # overall event record
                    slot=0,
                    can_slice_single=True,  # Will raise NotImplementedError (Run 2 not supported)
                    event_in_file=int(m.group(3)),  # Record within stream
                    lfn=stream_to_lfn.get(stream_num),  # Get LFN from stream mapping
                )

            m = rootioalg_datafile.search(line)
            if m:
                opened_lfns.add(m.group(1))

            m = rootiohandler_processed.search(line)
            if m:
                processed_lfns.add(m.group(1))

            m = fatal_pattern.search(line)
            if m:
                failed_event = int(m.group(1))

    # Determine suspect LFNs: XML statuses are the most reliable source,
    # fall back to log-based heuristics (e.g. for segfaults where the
    # XML input section may be missing or incomplete).
    xml_suspect_lfns: list[str] = []
    if input_file_statuses:
        xml_suspect_lfns = [s.lfn for s in input_file_statuses if s.status != "full"]

    log_unprocessed_lfns = list(opened_lfns - processed_lfns)

    # Prefer new-style (allows single-event slicing)
    if last_new is not None:
        last_new.unprocessed_lfns = log_unprocessed_lfns
        last_new.suspect_lfns = xml_suspect_lfns or log_unprocessed_lfns
        last_new.failed_event_overall = failed_event
        return last_new

    # Run 2-style: precise slicing not yet implemented
    if last_run2 is not None:
        raise NotImplementedError(
            "Run 2 style event messages detected (EventSelector). "
            "Precise slicing for Run 2 jobs is not yet implemented. "
            f"Last event found: {last_run2.event_number} (in-file: {last_run2.event_in_file})"
        )

    if last_old is not None:
        last_old.suspect_lfns = xml_suspect_lfns or (
            [last_old.lfn] if last_old.lfn else []
        )
        last_old.failed_event_overall = failed_event
        # Compute suspect in-file event when we have FATAL + RootIOAlg offset
        if failed_event is not None and last_old.event_in_file is not None:
            last_old.suspect_event_in_file = last_old.event_in_file + (
                failed_event - last_old.event_number
            )
        return last_old

    raise ValueError(
        f"No event processing messages found in {log_file}. "
        "The error may have occurred before any events were processed."
    )


def get_all_input_lfns_from_prodconf(prodconf_file: Path) -> list[str]:
    """Extract all input LFNs from a prodConf JSON file.

    Args:
        prodconf_file: Path to the prodConf JSON file.

    Returns:
        List of input file LFNs.
    """
    try:
        with open(prodconf_file) as f:
            prodconf = json.load(f)
        return prodconf.get("input", {}).get("files", [])
    except (json.JSONDecodeError, OSError):
        return []


def create_reproducer_prodconf(
    prodconf_file: Path,
    output_path: Path,
    *,
    input_files: list[str] | None = None,
    first_event: int = 0,
    n_of_events: int = -1,
    print_freq: int = 1,
    application_version: str | None = None,
    binary_tag: str | None = None,
) -> Path:
    """Create a modified prodConf JSON for the reproducer.

    Args:
        prodconf_file: Path to the original prodConf JSON.
        output_path: Where to write the modified prodConf.
        input_files: Override input files list. If None, keep original.
        first_event: First event number to process.
        n_of_events: Number of events (-1 for all).
        print_freq: Event print frequency (1 for debugging).
        application_version: Override application version. If None, keep original.
        binary_tag: Override binary tag. If None, keep original.

    Returns:
        Path to the written reproducer prodConf.

    Raises:
        FileNotFoundError: If the original prodConf doesn't exist.
    """
    with open(prodconf_file) as f:
        prodconf = json.load(f)

    if input_files is not None:
        prodconf["input"]["files"] = input_files

    prodconf["input"]["first_event_number"] = first_event
    prodconf["input"]["n_of_events"] = n_of_events

    if application_version is not None:
        prodconf["application"]["version"] = application_version
    if binary_tag is not None:
        prodconf["application"]["binary_tag"] = binary_tag

    if "extra_options" in prodconf.get("options", {}):
        prodconf["options"]["extra_options"]["print_freq"] = print_freq

    with open(output_path, "w") as f:
        json.dump(prodconf, f, indent=2)

    return output_path


def run_reproducer(
    job_dir: Path,
    reproducer_prodconf: Path,
    *,
    on_line: Callable[[str], None] | None = None,
) -> tuple[subprocess.CompletedProcess, Path]:
    """Run the reproducer in a clean copy of the job directory.

    Creates a ``.slice-work`` subdirectory with a copy of the job files,
    then runs ``lb-prod-run`` with the reproducer prodConf.

    Output is streamed line-by-line so that callers can show progress.
    The optional ``on_line`` callback is invoked for each stdout line.

    Args:
        job_dir: Original job directory.
        reproducer_prodconf: Path to the reproducer prodConf JSON.
        on_line: Optional callback invoked with each stdout line (stripped).

    Returns:
        Tuple of (CompletedProcess result, path to the slice working directory).
    """
    slice_dir = job_dir / ".slice-work"

    if slice_dir.exists():
        shutil.rmtree(slice_dir)

    shutil.copytree(
        job_dir,
        slice_dir,
        ignore=lambda d, _files: (
            [".slice-work", "minimal_reproducer", "reproducer"]
            if d == str(job_dir)
            else []
        ),
    )

    # Copy the reproducer prodconf into the slice directory
    slice_prodconf = slice_dir / reproducer_prodconf.name
    shutil.copy2(reproducer_prodconf, slice_prodconf)

    proc = subprocess.Popen(
        ["lb-prod-run", slice_prodconf.name],
        cwd=slice_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,  # line-buffered
    )

    # Read stderr in a background thread to avoid deadlocks
    stderr_lines: list[str] = []

    def _drain_stderr():
        assert proc.stderr is not None
        for line in proc.stderr:
            stderr_lines.append(line)

    stderr_thread = threading.Thread(target=_drain_stderr, daemon=True)
    stderr_thread.start()

    # Read stdout line-by-line, calling the progress callback
    stdout_lines: list[str] = []
    assert proc.stdout is not None
    for line in proc.stdout:
        stdout_lines.append(line)
        if on_line is not None:
            on_line(line.rstrip("\n"))

    proc.wait()
    stderr_thread.join(timeout=5)

    result = subprocess.CompletedProcess(
        args=proc.args,
        returncode=proc.returncode,
        stdout="".join(stdout_lines),
        stderr="".join(stderr_lines),
    )

    return result, slice_dir


def run_minimal_reproducer(
    minimal_reproducer_dir: Path,
    *,
    on_line: Callable[[str], None] | None = None,
) -> tuple[subprocess.CompletedProcess, Path]:
    """Run the minimal reproducer in a clean copy of the job directory.

    Creates a ``.slice-work`` subdirectory with a copy of the job files,
    then runs ``lb-prod-run`` with the reproducer prodConf.

    Output is streamed line-by-line so that callers can show progress.
    The optional ``on_line`` callback is invoked for each stdout line.

    Args:
        minimal_reproducer_dir: Directory containing the minimal reproducer.
        on_line: Optional callback invoked with each stdout line (stripped).

    Returns:
        Tuple of (CompletedProcess result, path to the slice working directory).
    """
    slice_dir = minimal_reproducer_dir / ".test-work"

    if slice_dir.exists():
        shutil.rmtree(slice_dir)

    shutil.copytree(
        minimal_reproducer_dir,
        slice_dir,
        ignore=lambda d, _files: (
            [".test-work", "minimal_reproducer", "reproducer"]
            if d == str(minimal_reproducer_dir)
            else []
        ),
    )

    proc = subprocess.Popen(
        ["bash", "run_reproducer.sh"],
        cwd=slice_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,  # line-buffered
    )

    # Read stderr in a background thread to avoid deadlocks
    stderr_lines: list[str] = []

    def _drain_stderr():
        assert proc.stderr is not None
        for line in proc.stderr:
            stderr_lines.append(line)

    stderr_thread = threading.Thread(target=_drain_stderr, daemon=True)
    stderr_thread.start()

    # Read stdout line-by-line, calling the progress callback
    stdout_lines: list[str] = []
    assert proc.stdout is not None
    for line in proc.stdout:
        stdout_lines.append(line)
        if on_line is not None:
            on_line(line.rstrip("\n"))

    proc.wait()
    stderr_thread.join(timeout=5)

    result = subprocess.CompletedProcess(
        args=proc.args,
        returncode=proc.returncode,
        stdout="".join(stdout_lines),
        stderr="".join(stderr_lines),
    )

    return result, slice_dir


def analyze_reproducer_output(
    stdout: str,
    stderr: str = "",
    *,
    first_event_offset: int = 0,
) -> ReproducerOutput:
    """Parse reproducer output for event information and failure messages.

    When ``first_event_offset`` is non-zero the reproducer was started
    with ``first_event_number > 0``, so event numbers in the log are
    relative to that starting point.  The offset is added to
    ``exact_problem_event`` so that it reflects the true in-file
    position.

    Searches both stdout and stderr for four patterns:

    1. **New style** (sets ``can_slice_single=True``)::

        HLTControlFlowMgr SUCCESS Processing event N in slot M

    2. **Old style** (in-file ↔ overall event mapping)::

        RootIOAlg SUCCESS Reading Event record N within LFN:..., M overall

    3. **Run 2 style** (in-file ↔ overall event mapping, sets ``can_slice_single=True``)::

        EventSelector SUCCESS Reading Event record M. Record number within stream 1: N

    4. **Bonus** (explicit failure with event number)::

        HLTControlFlowMgr FATAL Event N on slot M failed!

    The FATAL message typically appears on stderr.

    Args:
        stdout: The stdout from the reproducer run.
        stderr: The stderr from the reproducer run.

    Returns:
        ReproducerOutput with parsed event information.
    """
    new_style_pattern = re.compile(
        r"HLTControlFlowMgr\s+SUCCESS\s+Processing event\s+(\d+)\s+in slot\s+(\d+)"
    )
    old_style_pattern = re.compile(
        r"RootIOAlg\s+SUCCESS\s+Reading Event record\s+(\d+)\s+within\s+(LFN:[^,]+),\s+(\d+)\s+overall"
    )
    run2_style_pattern = re.compile(
        r"EventSelector\s+SUCCESS\s+Reading Event record\s+(\d+)\.\s+Record number within stream (\d+):\s+(\d+)"
    )
    run2_lfn_pattern = re.compile(
        r"EventSelector\s+INFO\s+Stream:EventSelector\.DataStreamTool_(\d+)\s+Def:DATAFILE='(LFN:[^']+)'"
    )
    fail_pattern = re.compile(r"Event\s+(\d+)\s+on\s+slot\s+\d+\s+failed")

    output = ReproducerOutput()
    # Map stream number to LFN for Run 2 style
    stream_to_lfn: dict[int, str] = {}

    combined = stdout + "\n" + stderr
    for line in combined.split("\n"):
        m = new_style_pattern.search(line)
        if m:
            output.last_processing_event = int(m.group(1))
            output.can_slice_single = True

        m = old_style_pattern.search(line)
        if m:
            output.last_event_in_file = int(m.group(1))
            output.last_event_lfn = m.group(2)
            output.last_event_overall = int(m.group(3))

        m = run2_lfn_pattern.search(line)
        if m:
            stream_num = int(m.group(1))
            lfn = m.group(2)
            stream_to_lfn[stream_num] = lfn

        m = run2_style_pattern.search(line)
        if m:
            stream_num = int(m.group(2))
            output.last_event_in_file = int(m.group(3))  # Record within stream
            output.last_event_overall = int(m.group(1))  # Overall event record
            output.last_event_lfn = stream_to_lfn.get(
                stream_num
            )  # Get LFN from stream mapping
            output.can_slice_single = True  # Can slice to single event

        m = fail_pattern.search(line)
        if m:
            output.failed_event_overall = int(m.group(1))

    # Calculate exact problem event
    if output.last_event_in_file is not None:
        if (
            output.failed_event_overall is not None
            and output.last_event_overall is not None
        ):
            # Use FATAL event number mapped back to in-file coordinates
            offset = output.last_event_overall - output.last_event_in_file
            output.exact_problem_event = output.failed_event_overall - offset
        elif (
            output.can_slice_single
            and output.last_processing_event is not None
            and output.last_event_overall is not None
        ):
            # New-style: last processing event is the one that crashed
            offset = output.last_event_overall - output.last_event_in_file
            output.exact_problem_event = output.last_processing_event - offset
        else:
            # Old-style only, no FATAL: last read event is best guess
            output.exact_problem_event = output.last_event_in_file
    elif output.can_slice_single and output.last_processing_event is not None:
        # New-style messages only, no RootIOAlg — use overall event number directly
        output.exact_problem_event = output.last_processing_event

    # Adjust for first_event_number: only needed when we have new-style
    # messages without old-style RootIOAlg records, because old-style
    # records already provide absolute in-file event positions.
    if (
        first_event_offset
        and output.exact_problem_event is not None
        and output.last_event_in_file is None
    ):
        output.exact_problem_event += first_event_offset - 1

    return output


def create_extraction_config(
    *,
    input_lfn: str,
    first_event: int,
    n_events: int = 1,
    output_file: Path,
    pool_catalog: Path | None = None,
    prodconf_file: Path | None = None,
) -> dict:
    """Create a YAML config dict for extracting events using GaudiConf.mergeDST.

    Args:
        input_lfn: The LFN of the input file.
        first_event: The first event number to extract.
        n_events: Number of events to extract (1 for single event).
        output_file: Path for the extracted output file.
        pool_catalog: Path to pool_xml_catalog.xml if available.
        prodconf_file: Path to original prodConf to infer parameters.

    Returns:
        Dictionary suitable for writing as YAML and passing to lbexec.
    """
    config = {
        "input_files": [input_lfn],
        "first_evt": first_event,
        "evt_max": n_events,
        "print_freq": max(1, first_event // 10),
        "output_file": str(output_file),
        "input_process": "Hlt2",
        "input_type": "ROOT",
        "input_raw_format": 0.5,
        "data_type": "Upgrade",
        "simulation": False,
        "geometry_version": "run3/2024.Q1.2-v00.00",
        "conditions_version": "master",
        "xml_file_catalog": (
            str(pool_catalog.resolve())
            if pool_catalog and pool_catalog.exists()
            else ""
        ),
        "output_type": "ROOT",
        "compression": "ZSTD:6",
        "root_ioalg_name": "RootIOAlgExt",
    }

    # Infer parameters from prodConf if available
    if prodconf_file and prodconf_file.exists():
        try:
            with open(prodconf_file) as f:
                prodconf = json.load(f)

            if prodconf.get("db_tags", {}).get("dddb_tag"):
                config["simulation"] = True

            app_name = prodconf.get("application", {}).get("name", "")
            if "Gauss" in app_name or "Sim" in app_name:
                config["simulation"] = True
        except (json.JSONDecodeError, OSError):
            pass

    return config


def extract_events(
    *,
    input_lfn: str,
    first_event: int,
    n_events: int = 1,
    extract_dir: Path,
    pool_catalog: Path | None = None,
    prodconf_file: Path | None = None,
    lhcb_version: str = "LHCb/v58r2",
) -> tuple[Path | None, subprocess.CompletedProcess]:
    """Extract events from an input file using GaudiConf.mergeDST.

    Args:
        input_lfn: The LFN of the input file.
        first_event: The first event number to extract.
        n_events: Number of events to extract (1 for single event).
        extract_dir: Working directory for extraction.
        pool_catalog: Path to pool_xml_catalog.xml if available.
        prodconf_file: Path to original prodConf to infer parameters.
        lhcb_version: LHCb version string for lb-run.

    Returns:
        Tuple of (path to extracted file or None, CompletedProcess).
    """
    extract_dir.mkdir(exist_ok=True, parents=True)

    # Copy pool catalog if available
    if pool_catalog and pool_catalog.exists():
        shutil.copy2(pool_catalog, extract_dir / "pool_xml_catalog.xml")

    if n_events == 1:
        output_file = extract_dir / f"event_{first_event}.dst"
    else:
        output_file = (
            extract_dir / f"events_{first_event}_to_{first_event + n_events}.dst"
        )

    config = create_extraction_config(
        input_lfn=input_lfn,
        first_event=first_event,
        n_events=n_events,
        output_file=output_file,
        pool_catalog=extract_dir / "pool_xml_catalog.xml" if pool_catalog else None,
        prodconf_file=prodconf_file,
    )

    config_path = extract_dir / "extract_event.yaml"
    with open(config_path, "w") as f:
        yaml.dump(config, f, default_flow_style=False, sort_keys=False)

    result = subprocess.run(
        ["lb-run", lhcb_version, "lbexec", "GaudiConf.mergeDST:dst", str(config_path)],
        cwd=extract_dir,
        capture_output=True,
        text=True,
    )

    # Save log
    log_path = extract_dir / "extraction.log"
    with open(log_path, "w") as f:
        f.write(
            f"STDOUT:\n{'=' * 80}\n{result.stdout}\n\nSTDERR:\n{'=' * 80}\n{result.stderr}\n"
        )

    if output_file.exists():
        return output_file, result
    return None, result


def extract_single_event(
    *,
    input_lfn: str,
    exact_event: int,
    extract_dir: Path,
    pool_catalog: Path | None = None,
    prodconf_file: Path | None = None,
    lhcb_version: str = "LHCb/v58r2",
) -> tuple[Path | None, subprocess.CompletedProcess]:
    """Extract a single event from an input file using GaudiConf.mergeDST.

    Thin wrapper around :func:`extract_events` with ``n_events=1``.
    """
    return extract_events(
        input_lfn=input_lfn,
        first_event=exact_event,
        n_events=1,
        extract_dir=extract_dir,
        pool_catalog=pool_catalog,
        prodconf_file=prodconf_file,
        lhcb_version=lhcb_version,
    )


def download_lfn_files(
    lfns: list[str],
    output_dir: Path,
) -> list[Path]:
    """Download LFN files from the grid using dirac-dms-get-file.

    Args:
        lfns: List of LFN strings (e.g. ``"LFN:/lhcb/..."``).
        output_dir: Directory to download files into.

    Returns:
        List of paths to downloaded files.

    Raises:
        FileNotFoundError: If ``dirac-dms-get-file`` is not available.
        RuntimeError: If the download command fails.
    """
    output_dir.mkdir(exist_ok=True, parents=True)

    result = subprocess.run(
        ["dirac-dms-get-file", "-D", str(output_dir), *lfns],
        capture_output=True,
        text=True,
    )

    if result.returncode != 0:
        raise RuntimeError(
            f"dirac-dms-get-file failed (exit code {result.returncode}):\n"
            f"{result.stderr}"
        )

    # Find downloaded files by matching LFN basenames
    downloaded: list[Path] = []
    for lfn in lfns:
        # LFN format: "LFN:/lhcb/..." or just "/lhcb/..."
        lfn_path = lfn.removeprefix("LFN:")
        basename = Path(lfn_path).name
        local_path = output_dir / basename
        if local_path.exists():
            downloaded.append(local_path)

    return downloaded


def find_data_packages(prodconf_file: Path) -> list[DataPackage]:
    """Find CVMFS data packages referenced in a prodConf.

    Parses ``application.data_pkgs`` entries like ``"AnalysisProductions.v1r4315"``
    and resolves them to CVMFS directories at
    ``/cvmfs/lhcb.cern.ch/lib/lhcb/DBASE/{PackageName}/{version}``.

    Args:
        prodconf_file: Path to the prodConf JSON.

    Returns:
        List of DataPackage entries with resolved CVMFS paths.
    """
    packages = []
    try:
        with open(prodconf_file) as f:
            prodconf = json.load(f)

        data_pkgs = prodconf.get("application", {}).get("data_pkgs", [])
        for pkg_spec in data_pkgs:
            if not isinstance(pkg_spec, str) or "." not in pkg_spec:
                continue

            # Format: "PackageName.version" e.g. "AnalysisProductions.v1r4315"
            name, version = pkg_spec.rsplit(".", 1)
            cvmfs_path = CVMFS_DBASE / name / version

            if cvmfs_path.exists():
                packages.append(
                    DataPackage(
                        name=name,
                        version=version,
                        cvmfs_path=cvmfs_path,
                    )
                )

    except (json.JSONDecodeError, OSError):
        pass

    return packages


def create_reproducer_package(
    *,
    output_dir: Path,
    job_dir: Path,
    failed_step: FailedStep,
    input_lfn: str | None,
    exact_event: int | None,
    local_input_files: list[Path] | None = None,
    prodconf_file: Path | None = None,
    data_packages: list[DataPackage] | None = None,
    lhcb_version: str = "LHCb/v58r2",
    job_id: str | None = None,
) -> Path:
    """Create a minimal reproducer directory with all necessary files.

    Args:
        output_dir: Directory for the reproducer package.
        job_dir: Original job directory.
        failed_step: Information about the failed step.
        input_lfn: The problematic input file LFN.
        exact_event: The exact problematic event number (if known).
        local_input_files: Local files to include (extracted events or downloads).
        prodconf_file: Path to the original prodConf.
        data_packages: List of DataPackage entries to include from CVMFS.
        lhcb_version: LHCb version string.

    Returns:
        Path to the reproducer package directory.
    """
    output_dir.mkdir(exist_ok=True, parents=True)

    # Copy input data files and determine reproducer input references
    if local_input_files:
        input_filenames = []
        for f in local_input_files:
            if f.exists():
                shutil.copy2(f, output_dir / f.name)
                input_filenames.append(f.name)
        inputs_for_reproducer = input_filenames or [input_lfn or "UNKNOWN"]
    else:
        inputs_for_reproducer = [input_lfn or "UNKNOWN"]

    # Create reproducer prodConf
    if prodconf_file and prodconf_file.exists():
        with open(prodconf_file) as f:
            prodconf = json.load(f)

        prodconf["input"]["files"] = inputs_for_reproducer
        if local_input_files:
            prodconf["input"]["first_event_number"] = 0
        prodconf["input"]["n_of_events"] = -1

        if "extra_options" in prodconf.get("options", {}):
            prodconf["options"]["extra_options"]["print_freq"] = 1

        # Remove data_pkgs since the files are now bundled locally
        if data_packages and "data_pkgs" in prodconf.get("application", {}):
            del prodconf["application"]["data_pkgs"]

        reproducer_json = output_dir / "reproducer.json"
        with open(reproducer_json, "w") as f:
            json.dump(prodconf, f, indent=2)

    # Copy data packages from CVMFS
    if data_packages:
        for dp in data_packages:
            dest = output_dir / dp.name / dp.version
            if dest.exists():
                shutil.rmtree(dest)
            shutil.copytree(dp.cvmfs_path, dest)

    # Create run script
    run_script = output_dir / "run_reproducer.sh"
    with open(run_script, "w") as f:
        f.write("#!/bin/bash\n")
        f.write("# Minimal reproducer for problematic event\n")
        f.write("# run with --interactive to get lb-run & lbexec command\n")
        f.write(f"# Original job: {job_dir}\n")
        f.write(f"# Failed step: {failed_step.step_id}\n")
        if exact_event is not None:
            f.write(f"# Problematic event: {exact_event}\n")
        if input_lfn:
            f.write(f"# Original file: {input_lfn}\n")
        f.write("\nset -e\n\n")
        # Add data package directories to PYTHONPATH
        if data_packages:
            f.write("# Add bundled data packages to PYTHONPATH\n")
            f.write('DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"\n')
            for dp in data_packages:
                if dp.name == "AnalysisProductions":
                    f.write(
                        f'export ANALYSIS_PRODUCTIONS_BASE="$DIR/{dp.name}/{dp.version}"\n'
                    )
                f.write(
                    f'export PYTHONPATH="$DIR/{dp.name}/{dp.version}:$PYTHONPATH"\n'
                )
            f.write("\n")
        f.write("echo 'Running minimal reproducer...'\n")
        if exact_event is not None:
            f.write(f"echo 'This will process the problematic event ({exact_event})'\n")
        f.write("echo ''\n\n")
        if prodconf_file and prodconf_file.exists():
            f.write("lb-prod-run reproducer.json $@\n")
        else:
            f.write("# No prodConf available - manual setup required\n")
            f.write(f"# lb-run {lhcb_version} <application> <options>\n")
    run_script.chmod(0o755)

    # Create README
    readme = output_dir / "README.txt"
    with open(readme, "w") as f:
        f.write("  ╔═╗╔═╗  LHCb Analysis Productions\n")
        f.write("  ╠═╣╠═╝  MINIMAL REPRODUCER PACKAGE\n")
        f.write("  ╩ ╩╩    ─────────────────────────────\n")
        f.write("\n")
        f.write("=" * 80 + "\n\n")
        f.write("Created by lb-ap make-reproducer\n")
        f.write("Address problems to DPA-WP2 Mattermost!\n\n")
        f.write("""
This directory contains everything needed to reproduce the error.
If you're able to fix it yourself consider making a test case from this if appropriate.
If you need help to fix the issue, please simplify your python code in $PWD/AnalysisProductions/v1rXXXX
to make it as minimal as possible while still producing the same error ideally aim to simplify it down
to a single python file which can be ran with:

    $ gaudirun.py my_reproducer.py ProdConfXXX.py
or  $ lbexec my_reproducer:main lbexec...yaml

""")
        f.write("FILES:\n")
        if local_input_files:
            for lf in local_input_files:
                f.write(f"  - {lf.name} - Input data file\n")
        if prodconf_file and prodconf_file.exists():
            f.write("  - reproducer.json - Job configuration\n")
        f.write("  - run_reproducer.sh - Script to run the reproducer\n")
        if data_packages:
            for dp in data_packages:
                f.write(f"  - {dp.name}/{dp.version}/ - Data package from CVMFS\n")
        f.write("  - README.txt - This file\n\n")
        f.write("ORIGINAL JOB INFO:\n")
        f.write(f"  Job directory: {job_dir}\n")
        f.write(f"  Job ID: {job_id}\n")
        f.write(f"  Failed step: {failed_step.step_id}\n")
        if input_lfn:
            f.write(f"  Original file: {input_lfn}\n")
        if exact_event is not None:
            f.write(f"  Event number: {exact_event}\n")
        f.write("\nTO RUN:\n")
        f.write("  ./run_reproducer.sh\n\n")
        f.write("TO RUN INTERACTIVELY:\n")
        if prodconf_file and prodconf_file.exists():
            f.write("  ./run_reproducer.sh --interactive\n\n")
        f.write("TO RUN MANUALLY (e.g. using stack environment):\n")
        if data_packages:
            f.write("# Add bundled data packages to PYTHONPATH\n")
            f.write('DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"\n')
            for dp in data_packages:
                if dp.name == "AnalysisProductions":
                    f.write(
                        f'export ANALYSIS_PRODUCTIONS_BASE="$DIR/{dp.name}/{dp.version}"\n'
                    )
                f.write(
                    f'export PYTHONPATH="$DIR/{dp.name}/{dp.version}:$PYTHONPATH"\n'
                )
            f.write("\n")
        f.write(
            "  ... lbexec {options file} {yaml file}  # you can find the command in the interactive mode\n\n"
        )

        f.write("SHARING:\n")
        f.write("  This directory is self-contained and can be shared for debugging.\n")
        if local_input_files:
            total_kb = sum(
                lf.stat().st_size / 1024 for lf in local_input_files if lf.exists()
            )
            f.write(f"  Input data size: ~{total_kb:.1f} KB\n")

    return output_dir


def create_tarball(source_dir: Path, tarball_path: Path) -> Path:
    """Create a gzipped tarball from a directory.

    Args:
        source_dir: Directory to package.
        tarball_path: Output tarball path.

    Returns:
        Path to the created tarball.
    """
    import tarfile

    tarball_dir_name = tarball_path.stem.replace(".tar", "")

    with tarfile.open(tarball_path, "w:gz") as tar:
        for item in sorted(source_dir.rglob("*")):
            if item.is_file():
                arcname = f"{tarball_dir_name}/{item.relative_to(source_dir)}"
                tar.add(item, arcname=arcname)

    return tarball_path



###############################################################################
# (c) Copyright 2026 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Rewrite CWL workflows to support custom step inputs with conditional execution.

This module rewrites generated CWL workflows before local execution so that
upstream steps can be skipped when a user provides custom input for a
downstream step. It uses CWL v1.2 features:

- ``when`` clauses to conditionally skip upstream steps
- ``pickValue: first_non_null`` to select between custom input and upstream output

This rewriting is done locally in LbAPLocal and does NOT affect the production
CWL generated by LbAPCommon. Production workflows remain clean and unchanged.

FIXME: This is brittle and should eventually be integrated into the CWL generation
logic in LbAPCommon.
"""

from __future__ import annotations

import logging
from pathlib import Path

from ruamel.yaml import YAML

logger = logging.getLogger(__name__)

CUSTOM_INPUT_ID = "custom-input"


def rewrite_cwl_for_custom_input(
    cwl_path: Path,
    custom_input_files: list[str],
    job_name: str,
) -> tuple[Path, Path]:
    """Rewrite a CWL workflow to support custom input for a specific step.

    When a user provides custom input files (e.g. pre-existing SPRUCE.DST),
    this function modifies the CWL workflow so that:

    1. Upstream steps that produce the input are skipped (via ``when``)
    2. The target step receives the custom input (via ``pickValue``)

    Args:
        cwl_path: Path to the generated CWL workflow file.
        custom_input_files: List of custom input file paths (local or LFN:).
        job_name: The job name from the CLI, used to identify the target step.

    Returns:
        Tuple of (modified_cwl_path, inputs_yaml_path).
    """
    yaml = YAML()
    yaml.default_flow_style = False
    yaml.width = 120

    with open(cwl_path) as f:
        cwl_data = yaml.load(f)

    # Find the target step and its containing transformation
    transform_idx, target_step_id = _find_target_step(cwl_data, job_name)
    transform_step = cwl_data["steps"][transform_idx]
    sub_workflow = transform_step["run"]

    logger.info(
        "Target step: %s (in transformation %d)",
        target_step_id,
        transform_idx + 1,
    )

    # Find upstream steps to skip
    upstream_ids = _find_upstream_steps(sub_workflow["steps"], target_step_id)
    logger.info("Upstream steps to skip: %s", upstream_ids)

    if not upstream_ids:
        logger.warning(
            "No upstream steps found for %s - custom input will replace "
            "the workflow input directly",
            target_step_id,
        )

    # --- 1. Add custom-input parameter at each nesting level ---

    # Main workflow inputs
    _add_optional_input(cwl_data["inputs"], CUSTOM_INPUT_ID)
    _make_input_optional(cwl_data["inputs"], "input-data")

    # Wire through to transformation step
    _add_step_input(transform_step["in"], CUSTOM_INPUT_ID, CUSTOM_INPUT_ID)

    # Transformation sub-workflow inputs
    _add_optional_input(sub_workflow["inputs"], CUSTOM_INPUT_ID)
    _make_input_optional(sub_workflow["inputs"], "input-data")

    # --- 2. Add `when` clauses to upstream steps ---

    for step in sub_workflow["steps"]:
        if step["id"] in upstream_ids:
            step["when"] = f"$(inputs['{CUSTOM_INPUT_ID}'] === null)"
            _add_step_input(step["in"], CUSTOM_INPUT_ID, CUSTOM_INPUT_ID)
            logger.info("Added when clause to step: %s", step["id"])

    # --- 3. Update target step's input-data to use pickValue ---

    for step in sub_workflow["steps"]:
        if step["id"] == target_step_id:
            for inp in step["in"]:
                if inp["id"] == "input-data":
                    original_source = inp["source"]
                    inp["source"] = [CUSTOM_INPUT_ID, original_source]
                    inp["pickValue"] = "first_non_null"
                    logger.info(
                        "Updated %s input-data: pickValue from [%s, %s]",
                        target_step_id,
                        CUSTOM_INPUT_ID,
                        original_source,
                    )
                    break

    # --- 4. Make outputs from skipped steps nullable ---

    for output in sub_workflow.get("outputs", []):
        output_source = output.get("outputSource")
        if output_source is None:
            continue
        # Check if outputSource references a skipped step
        source_str = output_source if isinstance(output_source, str) else None
        source_list = output_source if isinstance(output_source, list) else []

        sources_to_check = [source_str] if source_str else source_list
        for src in sources_to_check:
            if src and src.split("/")[0] in upstream_ids:
                _make_output_type_nullable(output)
                break

    # --- 5. Write modified CWL ---

    modified_cwl_path = cwl_path.parent / (cwl_path.stem + "_custom_input.cwl")
    with open(modified_cwl_path, "w") as f:
        yaml.dump(cwl_data, f)
    logger.info("Written modified CWL: %s", modified_cwl_path)

    # --- 6. Write inputs YAML ---

    inputs_yaml = _build_inputs_yaml(custom_input_files)
    inputs_yaml_path = cwl_path.parent / (cwl_path.stem + "_custom_inputs.yml")
    with open(inputs_yaml_path, "w") as f:
        yaml.dump(inputs_yaml, f)
    logger.info("Written inputs YAML: %s", inputs_yaml_path)

    return modified_cwl_path, inputs_yaml_path


def _find_target_step(cwl_data: dict, job_name: str) -> tuple[int, str]:
    """Find the target step within the workflow.

    Searches nested transformation sub-workflows for a step whose ID
    contains the job_name. Falls back to the last leaf step if no
    exact match is found.

    Returns:
        Tuple of (transformation_index, step_id).

    Raises:
        ValueError: If no suitable target step can be found.
    """
    # Normalise job_name for matching (replace commas/hyphens)
    job_name_normalised = job_name.replace(",", "_").replace("-", "_")

    for t_idx, transform_step in enumerate(cwl_data.get("steps", [])):
        sub_wf = transform_step.get("run", {})
        if sub_wf.get("class") != "Workflow":
            continue

        sub_steps = sub_wf.get("steps", [])
        if not sub_steps:
            continue

        # Try to match step ID against job_name
        for step in sub_steps:
            step_id = step["id"]
            step_id_normalised = step_id.replace("-", "_")
            if step_id_normalised.endswith(job_name_normalised):
                return t_idx, step_id

    # Fallback: use the last leaf step in the first transformation
    for t_idx, transform_step in enumerate(cwl_data.get("steps", [])):
        sub_wf = transform_step.get("run", {})
        if sub_wf.get("class") != "Workflow":
            continue

        sub_steps = sub_wf.get("steps", [])
        if not sub_steps:
            continue

        leaf_steps = _find_leaf_steps(sub_steps)
        if leaf_steps:
            target = leaf_steps[-1]
            logger.warning(
                "Could not match job_name '%s' to a step ID; "
                "falling back to last leaf step '%s'",
                job_name,
                target,
            )
            return t_idx, target

    raise ValueError(
        f"Could not find a target step for job_name '{job_name}' "
        f"in the CWL workflow"
    )


def _find_leaf_steps(steps: list[dict]) -> list[str]:
    """Find steps that no other step depends on (leaf nodes)."""
    all_ids = {step["id"] for step in steps}
    depended_on = set()

    for step in steps:
        for inp in step.get("in", []):
            source = inp.get("source", "")
            sources = [source] if isinstance(source, str) else source
            for src in sources:
                if "/" in src:
                    dep_id = src.split("/")[0]
                    if dep_id in all_ids:
                        depended_on.add(dep_id)

    return [s["id"] for s in steps if s["id"] not in depended_on]


def _find_upstream_steps(steps: list[dict], target_step_id: str) -> set[str]:
    """Find all steps upstream of the target (direct and transitive).

    Traces ``source`` references in step ``in`` bindings to build the
    dependency graph and walks it backwards from the target.
    """
    # Build dependency graph: step_id -> set of step_ids it depends on
    all_ids = {step["id"] for step in steps}
    deps: dict[str, set[str]] = {step["id"]: set() for step in steps}

    for step in steps:
        step_id = step["id"]
        for inp in step.get("in", []):
            source = inp.get("source", "")
            sources = [source] if isinstance(source, str) else source
            for src in sources:
                if "/" in src:
                    dep_id = src.split("/")[0]
                    if dep_id in all_ids:
                        deps[step_id].add(dep_id)

    # Walk backwards from target to find all ancestors
    ancestors: set[str] = set()
    to_visit = list(deps.get(target_step_id, set()))
    while to_visit:
        current = to_visit.pop()
        if current not in ancestors:
            ancestors.add(current)
            to_visit.extend(deps.get(current, set()))

    return ancestors


def _add_optional_input(inputs: list[dict], input_id: str) -> None:
    """Add an optional File[]? input to a workflow's inputs list."""
    # Check if already present
    for inp in inputs:
        if inp["id"] == input_id:
            return

    inputs.append(
        {
            "id": input_id,
            "doc": "Custom input files (skips upstream steps when provided)",
            "type": ["null", {"type": "array", "items": "File"}],
            "default": None,
        }
    )


def _make_input_optional(inputs: list[dict], input_id: str) -> None:
    """Change an input from required (File[]) to optional (File[]?)."""
    for inp in inputs:
        if inp["id"] == input_id:
            current_type = inp.get("type")
            if current_type == "File[]":
                inp["type"] = ["null", {"type": "array", "items": "File"}]
                inp["default"] = None
            break


def _add_step_input(step_ins: list[dict], input_id: str, source: str) -> None:
    """Add an input binding to a workflow step's ``in`` list."""
    for inp in step_ins:
        if inp["id"] == input_id:
            return
    step_ins.append({"id": input_id, "source": source})


def _make_output_type_nullable(output: dict) -> None:
    """Make an output type nullable if it isn't already."""
    current_type = output.get("type")
    if current_type is None:
        return

    # Already nullable (union type with null)
    if isinstance(current_type, list):
        return
    if isinstance(current_type, dict) and current_type.get("type") == "array":
        items = current_type.get("items", [])
        if isinstance(items, list) and "null" in items:
            return

    # Convert File[] to File[]?
    if current_type == "File[]":
        output["type"] = ["null", {"type": "array", "items": "File"}]


def _build_inputs_yaml(custom_input_files: list[str]) -> dict:
    """Build the CWL inputs YAML dict with custom input files."""
    file_entries = []
    for filepath in custom_input_files:
        if filepath.startswith("LFN:"):
            file_entries.append({"class": "File", "location": filepath})
        else:
            # Local file - use absolute path
            abs_path = str(Path(filepath).resolve())
            file_entries.append({"class": "File", "path": abs_path})

    return {CUSTOM_INPUT_ID: file_entries}



import os
from Configurables import (
    TupleToolDecayTreeFitter,
    TriggerTisTos,
    GaudiSequencer,
    CheckPV,
    TrackScaleState, TrackSmearState,
    DecayTreeTuple, TupleToolTISTOS, TupleToolGeometry,
    TupleToolDecay, TupleToolPid, TupleToolRecoStats,
    LoKi__Hybrid__TupleTool,
    TupleToolTagging,
    DaVinci,
    BTaggingTool,
    FilterDesktop
)
from FlavourTagging.Tunings import applyTuning as applyFTTuning

# Trigger lines
l0_lines = [
    'L0DiMuonDecision',
    'L0GlobalDecision'
    ]
hlt1_lines = [
     'Hlt1GlobalDecision',
     'Hlt1TrackMuonDecision',
     'Hlt1TrackMVADecision',
     'Hlt1TwoTrackMVADecision',
     'Hlt1B2PhiPhi_LTUNBDecision'
    ]
hlt2_lines = [
    'Hlt2Topo2BodyBBDTDecision',
    'Hlt2Topo3BodyBBDTDecision',
    'Hlt2Topo4BodyBBDTDecision',
    'Hlt2Topo2BodyDecision',
    'Hlt2Topo3BodyDecision',
    'Hlt2Topo4BodyDecision',
    'Hlt2IncPhiSidebandsDecision',
    'Hlt2PhiIncPhiDecision',
    'Hlt2PhiBs2PhiPhiDecision'
    ]
mtl = l0_lines + hlt1_lines + hlt2_lines


is_mc = DaVinci().Simulation
is_mdst = DaVinci().InputType.upper() == 'MDST'
oldline = False if DaVinci().DataType=='2018' else True
line = 'BetaSBs2DsPiDetachedLine' if oldline==False else 'B02DPiD2HHHBeauty2CharmLine'
tes_root = '/Event/AllStreams' if is_mc else '/Event/BhadronCompleteEvent'
if is_mdst:
    DaVinci().RootInTES = tes_root
location = 'Phys/{}/Particles'.format(line)
if not is_mdst:
    location = os.path.join(tes_root, location)
tuple = DecayTreeTuple('Bs2DsPi_Tuple')
tuple.Inputs = [location]
# Unit
SeqPhys = GaudiSequencer('SeqPhys')

if oldline:
    tuple.Decay = '[B0 -> ^(D- -> ^K- ^pi- ^K+) ^pi+]CC'
    tuple.Branches = {
        'B' : '[B0 ->  (D- ->  K-  pi-  K+)  pi+]CC',
        'Ds': '[B0 -> ^(D- ->  K-  pi-  K+)  pi+]CC',
        'hminus': '[B0 ->  (D- -> ^K-  pi-  K+)  pi+]CC',
        'piminus': '[B0 ->  (D- ->  K- ^pi-  K+)  pi+]CC',
        'hplus': '[B0 ->  (D- ->  K-  pi- ^K+)  pi+]CC',
        'piplus': '[B0 ->  (D- ->  K-  pi-  K+) ^pi+]CC'
        }
else:
    tuple.Decay = '[B_s0 -> ^(D_s- -> ^K- ^pi- ^K+) ^pi+]CC'
    tuple.Branches = {
        'B': '[B_s0 ->  (D_s- ->  K-  pi-  K+)  pi+]CC',
        'Ds': '[B_s0 -> ^(D_s- ->  K-  pi-  K+)  pi+]CC',
        'hminus': '[B_s0 ->  (D_s- -> ^K-  pi-  K+)  pi+]CC',
        'piminus': '[B_s0 ->  (D_s- ->  K- ^pi-  K+)  pi+]CC',
        'hplus': '[B_s0 ->  (D_s- ->  K-  pi- ^K+)  pi+]CC',
        'piplus': '[B_s0 ->  (D_s- ->  K-  pi-  K+) ^pi+]CC'
        }

# TupleToolDecay
names = ['B', 'Ds', 'hminus', 'piminus', 'hplus', 'piplus']
for particle in names:
   tuple.addTool(TupleToolDecay, name=particle)

# Tool list
tl= [
    'TupleToolAngles',
    'TupleToolKinematic',
    'TupleToolPid',
    'TupleToolTrackInfo',
    'TupleToolPrimaries',
    'TupleToolPropertime',
    'TupleToolEventInfo',
    'TupleToolRecoStats',
    'TupleToolGeometry',
    'TupleToolTISTOS'
    ]
if is_mc:
    tl += ['TupleToolMCTruth',
           'TupleToolMCBackgroundInfo']

tuple.ToolList += tl
# Refit
tuple.ReFitPVs = True

#Geo
TupleToolGeometry = TupleToolGeometry('TupleToolGeometry')
TupleToolGeometry.RefitPVs = True
TupleToolGeometry.Verbose = True
tuple.addTool(TupleToolGeometry)

# Loki
LoKi_B = LoKi__Hybrid__TupleTool('LoKi_B')
LoKi_B.Variables = {
    'ETA': 'ETA',
    'LOKI_FDCHI2': 'BPVVDCHI2',
    'LOKI_FDS': 'BPVDLS',
    'LOKI_DIRA': 'BPVDIRA',
    'LOKI_DTF_CTAU': 'DTF_CTAU( 0, True )',
    'LOKI_DTF_CTAUS': 'DTF_CTAUSIGNIFICANCE( 0, True )',
    'LOKI_DTF_CHI2NDOF': 'DTF_CHI2NDOF( True )',
    'LOKI_DTF_CTAUERR': 'DTF_CTAUERR( 0, True )',
    'LOKI_MASS_DsConstr': "DTF_FUN ( M , True , 'D_s-' )",
    'LOKI_DTF_VCHI2NDOF': 'DTF_FUN ( VFASPF(VCHI2/VDOF) , True )'
}
tuple.B.addTool(LoKi_B)
tuple.B.ToolList += ["LoKi::Hybrid::TupleTool/LoKi_B"]

# RecoStats to filling SpdMult, etc
tuple.addTool(TupleToolRecoStats, name='TupleToolRecoStats')
tuple.TupleToolRecoStats.Verbose = True

# Refit with mass constraint
PVFit = TupleToolDecayTreeFitter('PVFit')
PVFit.Verbose = True
PVFit.UpdateDaughters = True
PVFit.constrainToOriginVertex = True

PVFitDs = TupleToolDecayTreeFitter('PVFitDs')
PVFitDs.Verbose = True
PVFitDs.UpdateDaughters = True
PVFitDs.constrainToOriginVertex = True
if oldline:
   PVFitDs.Substitutions = {
       'B0 -> ^(D- -> K- pi- K+) pi+': 'D_s-',
       'B~0-> ^(D+ -> K+ pi+ K-) pi-': 'D_s+'
       }
PVFitDs.daughtersToConstrain = ['D_s-']
tuple.B.addTool(PVFit)
tuple.B.addTool(PVFitDs)
tuple.B.ToolList += ["TupleToolDecayTreeFitter/PVFit"]
tuple.B.ToolList += ["TupleToolDecayTreeFitter/PVFitDs"]

# Trigger
TisTos = TupleToolTISTOS('TisTos')
TisTos.Verbose = True
TisTos.TriggerList = mtl

# Change TOSFracMu for Ds, B
tuple.B.addTool(TisTos)
tuple.B.TisTos.addTool(TriggerTisTos())
tuple.B.ToolList += ['TupleToolTISTOS/TisTos']
tuple.B.TisTos.TriggerTisTos.TOSFracMuon = 0
tuple.B.TisTos.TriggerTisTos.TOSFracEcal = 0
tuple.B.TisTos.TriggerTisTos.TOSFracHcal = 0

# specific tagging tuple tool
PTagging = TupleToolTagging('PTagging')
if is_mdst:
    PTagging.UseFTfromDST = True
else:
    PTagging.Verbose = True
    btagtool = PTagging.addTool(BTaggingTool, name='MyBTaggingTool')

    applyFTTuning(btagtool, tuning_version='Summer2017Optimisation_v4_Run2')
    PTagging.TaggingToolName = btagtool.getFullName()
tuple.B.addTool(PTagging)
tuple.B.ToolList += ['TupleToolTagging/PTagging']

# primary vertex selection
checkpv = CheckPV()
# Require that there are valid PVs after refitting, as required by
# TupleToolPropertime
checkrefitpv = FilterDesktop('CheckRefittedPVs_{}'.format(tuple.name()),
                             Code='BPVVALID()',
                             Inputs=tuple.Inputs,
                             ReFitPVs=True)

# apply momentum scaling for data and momentum smearing for MC
scale_or_smear = TrackSmearState('TrackSmearState') if is_mc else TrackScaleState('TrackScaleState')

SeqPhys.Members += [checkpv, checkrefitpv, scale_or_smear, tuple]
DaVinci().UserAlgorithms = [SeqPhys]



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json
import os
import shutil
import subprocess
from os.path import dirname, join
from pathlib import Path

import pytest
import yaml

EXAMPLE_REPO_ROOT = join(dirname(__file__), "data", "data-pkg-repo")
B02DKPi_versions = [
    "v0r0p2180615",
    "v0r0p2297458",
    "v0r0p2358939",
    "v0r0p2437974",
    "v0r0p2518507",
    "v0r0p2552694",
    "v0r0p2752886",
    "v0r0p2794507",
    "v0r0p2847846",
    "v0r0p2970193",
    "v0r0p3234056",
    "v0r0p3675801",
]


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def empty_data_pkg_repo(tmpdir):
    cwd = os.getcwd()
    os.chdir(tmpdir)
    yield tmpdir
    os.chdir(cwd)


@pytest.fixture
def data_pkg_repo(empty_data_pkg_repo):
    shutil.copytree(
        EXAMPLE_REPO_ROOT, empty_data_pkg_repo, symlinks=True, dirs_exist_ok=True
    )
    yield empty_data_pkg_repo


@pytest.fixture
def with_proxy(mocker):
    from subprocess import check_call as original_check_call

    def my_check_call(args, *more_args, **kwargs):
        if args == ["lb-dirac", "dirac-proxy-info", "--checkvalid"]:
            return
        else:
            return original_check_call(args, *more_args, **kwargs)

    mocker.patch("subprocess.check_call", new=my_check_call)


@pytest.fixture
def without_proxy(mocker):
    mocker.patch(
        "subprocess.check_call", side_effect=subprocess.CalledProcessError(1, "mocked")
    )


@pytest.fixture
def with_data(mocker, monkeypatch):
    from subprocess import run as original_run

    monkeypatch.setenv("LBRUN_BINDS", Path(__file__).parent)

    data_fn1 = (
        Path(__file__).parent / "test-input-dst" / "00070793_00000368_7.AllStreams.dst"
    )
    data_fn2 = (
        Path(__file__).parent / "test-input-dst" / "00070767_00000368_7.AllStreams.dst"
    )

    pool_catalog_xml = (
        '<?xml version="1.0" encoding="UTF-8" standalone="no" ?>\n'
        "<!-- Edited By PoolXMLCatalog.py -->\n"
        '<!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">\n'
        "<POOLFILECATALOG>\n"
        '<File ID="980B119E-9609-E811-A58E-02163E016452">\n'
        "<physical>\n"
        f'    <pfn filetype="ROOTAll" name="{data_fn1}" se="CI-DST"/>\n'
        "</physical>\n"
        "<logical>\n"
        '    <lfn name="/lhcb/MC/2016/ALLSTREAMS.DST/00070793/0000/00070793_00000368_7.AllStreams.dst"/>\n'
        "</logical>\n"
        "</File>\n"
        '<File ID="043C600A-8F09-E811-B65B-FA163E9FB5C5">\n'
        "<physical>\n"
        f'    <pfn filetype="ROOT_All" name="{data_fn2}" se="CI-DST"/>\n'
        "</physical>\n"
        "<logical>\n"
        '    <lfn name="/lhcb/MC/2016/ALLSTREAMS.DST/00070767/0000/00070767_00000368_7.AllStreams.dst"/>\n'
        "</logical>\n"
        "</File>\n"
        "</POOLFILECATALOG>\n"
    )

    def my_run(args, *more_args, **kwargs):
        if args[:2] == ["lb-dirac", "dirac-bookkeeping-genXMLCatalog"]:
            with open(join(kwargs["cwd"], "pool_xml_catalog.xml"), "wt") as fp:
                fp.write(pool_catalog_xml)
            return subprocess.CompletedProcess([], 0, "")
        else:
            return original_run(args, *more_args, **kwargs)

    mocker.patch("subprocess.run", new=my_run)

    yield pool_catalog_xml


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["authorization", "set-cookie"],
        "ignore_hosts": ["auth.cern.ch", "s3.cern.ch"],
    }


def monkeypatch_productions_to_versions(
    working_group: str, production_name: str
) -> list:
    return B02DKPi_versions


@pytest.fixture(autouse=True)
def production_to_versions_patched(monkeypatch):
    import LbAPLocal.cli
    import LbAPLocal.cli.legacy

    monkeypatch.setattr(
        LbAPLocal.cli, "production_to_versions", monkeypatch_productions_to_versions
    )
    monkeypatch.setattr(
        LbAPLocal.cli.legacy,
        "production_to_versions",
        monkeypatch_productions_to_versions,
    )


@pytest.fixture
def fake_global_git_config(tmp_path, monkeypatch):
    """Create a fake global git config file."""
    git_config = tmp_path / "gitconfig"
    lines = [
        "[user]",
        "    name = Test User",
        "    email = test.user@domain.invalid",
    ]
    git_config.write_text("\n".join(lines))
    monkeypatch.setenv("GIT_CONFIG_GLOBAL", str(git_config))


# ============================================================================
# CWL test fixtures
# ============================================================================

from fixture_loader import TEST_DST_DIR, get_fixture  # noqa: E402


@pytest.fixture
def cwl_fixture_set(request):
    """Return FixtureSet for the current test. Override via indirect parametrize."""
    production_name = getattr(request, "param", "executabletests")
    return get_fixture(production_name)


@pytest.fixture
def mock_dirac_calls(monkeypatch, cwl_fixture_set):
    """Mock DIRAC subprocess calls for CWL testing.

    Data-driven from tests/fixtures/ — loads query_dirac results and
    DST file mappings from fixture files.

    Mocks two layers:
    1. LbAPCommon's query_dirac() - returns fixture-loaded results
    2. subprocess.Popen for dirac-cwl-run - pre-generates inputs and replica map,
       strips --force-regenerate and --n-lfns flags so existing files are used
    """
    fixture_set = cwl_fixture_set

    # Verify test DST files exist for all jobs
    for job_name in fixture_set.job_names:
        dst_path = fixture_set.get_dst_path(job_name)
        if not dst_path.is_file():
            pytest.skip(f"Test data file not found: {dst_path}")

    # Bind test data directory into lb-run container
    monkeypatch.setenv("LBRUN_BINDS", str(TEST_DST_DIR))

    # ---- Mock #1: LbAPCommon.__main__.query_dirac ----
    async def fake_query_dirac(
        sem, query_script, name, job_data, output_path, server_credentials
    ):
        """Return fixture-loaded query_dirac results without calling lb-dirac."""
        async with sem:
            result = fixture_set.get_query_result(name)
            output_path.write_text(json.dumps(result))
            return name, job_data, result

    import LbAPCommon.__main__

    monkeypatch.setattr(LbAPCommon.__main__, "query_dirac", fake_query_dirac)

    # ---- Mock #2: subprocess.Popen for dirac-cwl-run ----
    original_popen = subprocess.Popen

    def patched_popen(args, *a, **kw):
        if isinstance(args, list) and args and "dirac-cwl-run" in args[0]:
            # Find the CWL workflow file
            cwl_file = None
            for arg in args:
                if str(arg).endswith(".cwl"):
                    cwl_file = Path(arg)
                    break

            if cwl_file and cwl_file.exists():
                # Match job name from the CWL filename stem
                stem = cwl_file.stem
                matched_job = None
                for job_name in fixture_set.job_names:
                    if job_name in stem:
                        matched_job = job_name
                        break

                if matched_job:
                    lfn = fixture_set.get_lfn(matched_job)
                    local_file = fixture_set.get_dst_path(matched_job)

                    # Write inputs YAML
                    inputs_path = cwl_file.parent / f"{cwl_file.stem}-inputs.yml"
                    inputs_data = {
                        "input-data": [{"class": "File", "location": f"LFN:{lfn}"}]
                    }
                    with open(inputs_path, "w") as f:
                        yaml.dump(
                            inputs_data, f, default_flow_style=False, sort_keys=False
                        )

                    # Write replica map JSON
                    replica_map_path = (
                        cwl_file.parent / f"{cwl_file.stem}-replica-map.json"
                    )
                    replica_map = {
                        lfn: {
                            "replicas": [
                                {
                                    "url": local_file.resolve().as_uri(),
                                    "se": "CI-DST",
                                }
                            ]
                        }
                    }
                    with open(replica_map_path, "w") as f:
                        json.dump(replica_map, f, indent=2)

            # Strip --force-regenerate and --n-lfns from the command
            # so dirac-cwl-run uses the pre-generated files
            new_args = []
            skip_next = False
            for arg in args:
                if skip_next:
                    skip_next = False
                    continue
                if arg == "--force-regenerate":
                    continue
                if arg == "--n-lfns":
                    skip_next = True
                    continue
                if arg == "--pick-smallest-lfn":
                    continue
                new_args.append(arg)
            args = new_args

        return original_popen(args, *a, **kw)

    monkeypatch.setattr(subprocess, "Popen", patched_popen)

    return fixture_set



import os
import sys

if 'ANALYSIS_PRODUCTIONS_BASE' in os.environ:
    sys.path.append(os.path.join(os.environ['ANALYSIS_PRODUCTIONS_BASE'], 'RDs'))
else:
    sys.path.append(os.getcwd())

from Configurables import DaVinci
from helpers import signal_tuple_maker


DaVinci().Simulation = True 
DaVinci().InputType = 'DST'
DaVinci().DataType = '2011'
DaVinci().PrintFreq = 5000
MagnetPolarity = 'MD'
DaVinci().EvtMax = -1

DaVinci().Lumi = not DaVinci().Simulation
DaVinci().TupleFile = 'Bsntuple_MC.root'
DaVinci().DDDBtag = "dddb-20170721-1"
CondDB_tag  = "sim-20160614-1-vc"
if MagnetPolarity == 'MU':
    DaVinci().CondDBtag = CondDB_tag + "-mu100"
elif MagnetPolarity == 'MD':
    DaVinci().CondDBtag = CondDB_tag + "-md100"

signal_tuple_maker.addNtupleToDaVinci(DaVinci().Simulation, 'RXcHad.Strip', 'Bs2DsTauNuForB2XTauNu')



###############################################################################
# (c) Copyright 2026 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Load CWL test fixture data from tests/fixtures/ directory.

Fixtures are organized by production name, with a manifest.yaml
that ties jobs to DST files and LFNs. Each job has a query_result.json
containing the captured output of LbAPCommon's query_dirac().

Usage:
    from fixture_loader import get_fixture, discover_fixtures

    # Load a specific fixture set
    fs = get_fixture("executabletests")
    result = fs.get_query_result("2016_MagDown_PromptMC_D02KK")

    # Auto-discover all fixtures
    all_fixtures = discover_fixtures()
"""

from __future__ import annotations

import json
from pathlib import Path

import yaml

FIXTURES_DIR = Path(__file__).parent / "fixtures"
TEST_DST_DIR = Path(__file__).parent / "test-input-dst"


class FixtureSet:
    """Loaded fixture data for one production."""

    def __init__(self, production_name: str):
        self._production_name = production_name
        self._fixture_dir = FIXTURES_DIR / production_name

        manifest_path = self._fixture_dir / "manifest.yaml"
        with open(manifest_path) as f:
            self._manifest = yaml.safe_load(f)

        self._query_results: dict[str, dict] = {}

    @property
    def production_name(self) -> str:
        return self._production_name

    @property
    def ap_repo(self) -> str:
        return self._manifest.get("ap_repo", "data-pkg-repo")

    @property
    def job_names(self) -> list[str]:
        return list(self._manifest["jobs"].keys())

    def _resolve_job(self, job_name: str) -> str:
        """Resolve same_as references to the actual job name."""
        config = self._manifest["jobs"][job_name]
        if "same_as" in config:
            return config["same_as"]
        return job_name

    def get_job_config(self, job_name: str) -> dict:
        """Get manifest config for a job, resolving same_as references."""
        actual = self._resolve_job(job_name)
        return self._manifest["jobs"][actual]

    def get_query_result(self, job_name: str) -> dict:
        """Load the query_result.json for a job."""
        actual = self._resolve_job(job_name)
        if actual not in self._query_results:
            result_path = self._fixture_dir / actual / "query_result.json"
            with open(result_path) as f:
                self._query_results[actual] = json.load(f)
        return self._query_results[actual]

    def get_dst_path(self, job_name: str) -> Path:
        """Get absolute path to the local DST file for a job."""
        config = self.get_job_config(job_name)
        return TEST_DST_DIR / config["dst_file"]

    def get_lfn(self, job_name: str) -> str:
        """Get the LFN that maps to the local DST for a job."""
        config = self.get_job_config(job_name)
        return config["lfn"]


def get_fixture(production_name: str) -> FixtureSet:
    """Get a specific fixture set by production name."""
    return FixtureSet(production_name)


def discover_fixtures() -> dict[str, FixtureSet]:
    """Discover all available fixture sets by scanning tests/fixtures/."""
    fixtures = {}
    if FIXTURES_DIR.exists():
        for manifest_path in sorted(FIXTURES_DIR.glob("*/manifest.yaml")):
            production_name = manifest_path.parent.name
            fixtures[production_name] = FixtureSet(production_name)
    return fixtures



"""Set up the main and tuple sequences.

The MainSeq sequence is set as the sole member of DaVinci().UserAlgorithms.
It contains:

1. CheckPV algorithm.
2. TrackSmearState algorithm if MC, else TrackScaleState.
3. Tuple sequence, as returned by helpers.tuple_maker.tuple_sequence.

Ntuple algorithms should be added to the TupleSeq sequence, which runs in
non-lazy OR mode.

The DaVinci().Simulation property must be set before this options file is run.
"""
import os
import sys
sys.path.append(os.path.join(os.environ['ANALYSIS_PRODUCTIONS_BASE'], 'BsToJpsiPhi'))

from Configurables import (
    CheckPV,
    DaVinci,
    GaudiSequencer,
    TrackScaleState,
    TrackSmearState
)

from helpers import tuple_maker

if DaVinci().Simulation:
    track_alg = TrackSmearState('TrackSmearState')
else:
    track_alg = TrackScaleState('TrackScaleState')

main_seq = GaudiSequencer('MainSeq')
main_seq.Members = [
    CheckPV(),
    track_alg,
    tuple_maker.tuple_sequence()
]

DaVinci().UserAlgorithms = [main_seq]



###############################################################################
# (c) Copyright 2020-2024 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
CLI package for LbAPLocal.

Contains all command-line interface implementations:
- main: Enhanced typer/rich-based CLI (primary interface)
- legacy: Original click-based CLI (deprecated)
- wizards: Interactive wizard utilities
"""

from apd.authentication import get_auth_headers

from ..utils import production_to_versions
from .legacy import main as legacy_main
from .main import app

__all__ = ["app", "legacy_main", "production_to_versions", "get_auth_headers"]



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from glob import glob
from os.path import basename, dirname, join

import pytest
from click.testing import CliRunner

from LbAPLocal.cli import legacy_main as main

log_paths = glob(join(dirname(__file__), "data", "example-logs", "*.log"))
assert len(log_paths) > 5


@pytest.mark.parametrize("log_path", log_paths)
def test_parse_logs(log_path):
    runner = CliRunner()
    result = runner.invoke(main, ["parse-log", log_path])

    if basename(log_path).startswith("good-"):
        assert result.exit_code == 0, result.stdout
    elif basename(log_path).startswith("error-"):
        assert result.exit_code == 1, result.stdout
    else:
        raise NotImplementedError(log_path)



from Gaudi.Configuration       import *
from GaudiKernel.SystemOfUnits import *
from Configurables import LoKi__Hybrid__TupleTool,DaVinci

def getLoKiToolsDictionary():

    reldict = {'RELINFO':{}}
    myvars = {
        #add RelInfoConeVariables
        'cone' : {'vars' : ['CONEANGLE','CONEMULT','CONEPTASYM'],
                  'locations' : ['TrackIsoInfoH1','TrackIsoInfoH2','TrackIsoInfoL1','TrackIsoInfoL2'] },
        #add RelInfoTrackIsolationBDT
        'isoBDT' : {'vars' : ['TRKISOBDTFIRSTVALUE','TRKISOBDTSECONDVALUE','TRKISOBDTTHIRDVALUE'],
                    'locations' : ['TrackIsoBDTInfoH1','TrackIsoBDTInfoH2','TrackIsoBDTInfoL1','TrackIsoBDTInfoL2'] } ,
        #add RelInfoConeIsolation
        'coneIso' : {'vars' : ['CC_ANGLE','CC_MULT','CC_PTASYM','CC_SPT','CC_PX','CC_PY','CC_PZ','CC_PXASYM','CC_PYASYM','CC_PZASYM','CC_DELTAETA','CC_DELTAPHI',
                               'NC_ANGLE','NC_MULT','NC_PTASYM','NC_SPT','NC_PX','NC_PY','NC_PZ','NC_PXASYM','NC_PYASYM','NC_PZASYM','NC_DELTAETA','NC_DELTAPHI',
                               'CC_MAXPT_PT','CC_MAXPT_PX','CC_MAXPT_PY','CC_MAXPT_PZ','NC_MAXPT_PT','NC_MAXPT_PX','NC_MAXPT_PY','NC_MAXPT_PZ',
                               'CC_IT','NC_IT','CCNC_IT'],
                    'locations' : ['ConeIsoInfoH1','ConeIsoInfoH2','ConeIsoInfoL1','ConeIsoInfoL2'] },
        #add RelInfoBs2MuMuTrackIsolations
        'Bs2MuMuIso' : {'vars' : ['BSMUMUTRACKPLUSISO', 'BSMUMUTRACKPLUSISOTWO', 'BSMUMUTRACKID', 'BSMUMUTRACKTOPID'],
                    'locations' : ['TrackIsoBs2MMInfoH1', 'TrackIsoBs2MMInfoH2', 'TrackIsoBs2MMInfoL1', 'TrackIsoBs2MMInfoL2'] },
         }

    for name,vardict in myvars.iteritems() :
        for var in vardict['vars'] :
            for loc in vardict['locations'] :
                reldict['RELINFO'][loc+"_"+var] = {'varName':var,'Location':loc,'Default':-1.}

    #add RelInfoVertexIsolation
    for v in [ 'VTXISONUMVTX', 'VTXISODCHI2ONETRACK', 'VTXISODCHI2MASSONETRACK', 'VTXISODCHI2TWOTRACK', 'VTXISODCHI2MASSTWOTRACK']:
        reldict['RELINFO']['VtxIso_'+v] = {'varName':v,'Location':'VertexIsoInfo','Default':-1.}

    #add RelInfoVertexIsolationBDT
    for v in [ 'VTXISOBDTHARDFIRSTVALUE', 'VTXISOBDTHARDSECONDVALUE', 'VTXISOBDTHARDTHIRDVALUE',]:
        reldict['RELINFO']['VtxIsoBDT_'+v] = {'varName':v,'Location':'VertexIsoBDTInfo','Default':-1.}
    
    return reldict

def getLoKiTool(name,location,is_mdst) :

    lokiDict = getLoKiToolsDictionary()
    LoKi_Tool = LoKi__Hybrid__TupleTool('LoKi_Tool'+name)

    if is_mdst: # If MDST, start from top path (IGNORE RootInTES for RelatedInfo)
        RootInTES = DaVinci().RootInTES
        location =  RootInTES + '/' + location
    for name, args in lokiDict['RELINFO'].iteritems(): 
        lokipath = location.replace('Particles',args['Location'])
 
        print(location, lokipath)
        LoKi_Tool.Variables[name] = "RELINFO('%s','%s',%f)"%(lokipath,args['varName'],args['Default'])

    return LoKi_Tool



"""Set up the main and tuple sequences.

The MainSeq sequence is set as the sole member of DaVinci().UserAlgorithms.
Ntuple algorithms should be added to the TupleSeq sequence, which runs in
non-lazy OR mode.

The DaVinci().Simulation property must be set before this options file is run.
"""
from Configurables import (
    DaVinci,
    GaudiSequencer,
)

from cbtesting.helpers import tuple_maker

main_seq = GaudiSequencer('MainSeq')
main_seq.Members = [
    tuple_maker.tuple_sequence(),
]

DaVinci().UserAlgorithms = [main_seq]

# ONLY CONFIG NEEDED
tuple_names = [
     'Lb2JpsiL_eeTuple',
     'Lb2JpsiL_mmTuple',
     'Lb2LemuTuple',
     'Lb2Lee_SSTuple',
     'Lb2Lmm_SSTuple',
     'Lb2Lemu_SSTuple',
     'Bu2JpsiK_eeTuple',
     'Bu2JpsiK_mmTuple',
     'Bd2JpsiKs_eeTuple',
     'Bd2JpsiKs_mmTuple',
    ]

for tuple_name in tuple_names:
    use_upstream_electrons = False
    if 'ee' in tuple_name and not 'SS' in tuple_name: 
        use_upstream_electrons = True
    tuple_maker.tuple_maker( tuple_name, upstream_electrons=use_upstream_electrons) 






#!/usr/bin/env python
###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".  #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Re-slice an existing local DST file to a smaller event range.

Use this after capture-fixture to narrow down the DST to only the events
that make the CWL test workflows produce output.

On first run the original file is backed up to <file>.bak and the sliced
output replaces it.  On subsequent runs the .bak (unsliced original) is
used as the source so you can refine first_evt / evt_max without losing
the full file.

Prerequisites:
    - CVMFS available (for lb-run)

Usage:
    pixi run reslice-fixture <dst-file> --first-evt 50 --evt-max 5

Example:
    pixi run reslice-fixture tests/test-input-dst/00070793_00000685_7.AllStreams_slice.dst --first-evt 15 --evt-max 20
"""

from __future__ import annotations

import argparse
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import yaml


def main():
    parser = argparse.ArgumentParser(
        description="Re-slice an existing local DST to a smaller event range",
    )
    parser.add_argument(
        "dst_file", type=Path, help="Path to the local DST file to re-slice"
    )
    parser.add_argument(
        "--first-evt",
        type=int,
        default=0,
        help="First event index to keep (default: 0)",
    )
    parser.add_argument(
        "--evt-max",
        type=int,
        default=10,
        help="Number of events to keep (default: 10)",
    )
    args = parser.parse_args()

    dst_file = args.dst_file.resolve()
    bak_file = dst_file.with_suffix(dst_file.suffix + ".bak")

    # Determine the source file to slice from:
    # - If a .bak exists, always slice from it (the unsliced original)
    # - Otherwise slice from the dst_file itself (first run)
    if bak_file.exists():
        source_file = bak_file
        print(f"Backup:    {bak_file} (using existing backup as source)")
    elif dst_file.exists():
        source_file = dst_file
    else:
        print(f"ERROR: DST file not found: {dst_file}", file=sys.stderr)
        sys.exit(1)

    original_size = source_file.stat().st_size
    print(f"Input:     {source_file} ({original_size / 1024:.0f} KB)")
    print(f"Range:     first_evt={args.first_evt}, evt_max={args.evt_max}")

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        output_file = tmpdir / dst_file.name

        slice_config = {
            "input_files": [str(source_file)],
            "first_evt": args.first_evt,
            "evt_max": args.evt_max,
            "output_file": str(output_file),
            "simulation": True,
            "input_process": "Hlt2",
            "input_type": "ROOT",
            "input_raw_format": 0.5,
            "data_type": "Upgrade",
            "geometry_version": "run3/2024.Q1.2-v00.00",
            "conditions_version": "master",
            "output_type": "ROOT",
            "compression": "ZSTD:6",
            "root_ioalg_name": "RootIOAlgExt",
        }

        config_path = tmpdir / "reslice_config.yaml"
        with open(config_path, "w") as f:
            yaml.dump(slice_config, f, default_flow_style=False, sort_keys=False)

        print(
            f"Running:   lb-run LHCb/v58r2 lbexec GaudiConf.mergeDST:dst {config_path}"
        )
        result = subprocess.run(
            [
                "lb-run",
                "LHCb/v58r2",
                "lbexec",
                "GaudiConf.mergeDST:dst",
                str(config_path),
            ],
        )
        if result.returncode != 0:
            print(
                f"ERROR: Slice failed with exit code {result.returncode}",
                file=sys.stderr,
            )
            sys.exit(1)

        if not output_file.exists():
            print(f"ERROR: Output file not created: {output_file}", file=sys.stderr)
            sys.exit(1)

        new_size = output_file.stat().st_size

        # Back up the original on first run (before any slicing)
        if not bak_file.exists():
            shutil.move(str(dst_file), str(bak_file))
            print(f"Backup:    {dst_file} -> {bak_file}")

        shutil.move(str(output_file), str(dst_file))

    print(f"Output:    {dst_file} ({new_size / 1024:.0f} KB)")
    print(
        f"Reduction: {original_size / 1024:.0f} KB -> {new_size / 1024:.0f} KB "
        f"({100 * (1 - new_size / original_size):.0f}% smaller)"
    )


if __name__ == "__main__":
    main()



from Configurables import DaVinci

from cbtesting.helpers import stripping

stripping_lines = [
                'StrippingBu2LLK_meLine',
                'StrippingBu2LLK_meSSLine',
                'StrippingBu2LLK_eeLine',
                'StrippingBu2LLK_eeLine2',
                'StrippingBu2LLK_eeLine3',
                'StrippingBu2LLK_eeLine4',
                'StrippingBu2LLK_mmLine',
                'StrippingBu2LLK_mmSSLine',
                'StrippingBu2LLK_eeSSLine2',
                ]
data_type = DaVinci().DataType

seq = stripping.stripping(data_type, stripping_lines)
DaVinci().UserAlgorithms = [seq]



from Configurables import (
    DaVinci,
    DecayTreeTuple
)
from DecayTreeTuple.Configuration import *

# Stream and stripping line we want to use
stream = 'AllStreams'
line = 'D2hhPromptDst2D2KKLine'

# Create an ntuple to capture D*+ decays from the StrippingLine line
dtt = DecayTreeTuple('TupleDstToD0pi_D0ToKK')
dtt.Inputs = ['/Event/{0}/Phys/{1}/Particles'.format(stream, line)]
dtt.Decay = '[D*(2010)+ -> ^(D0 -> ^K- ^K+) ^pi+]CC'

# add a kinematic fitter
dtt.addBranches({
    'Dstar': '[D*(2010)+ -> (D0 -> K- K+) pi+]CC',
})
dtt.Dstar.addTupleTool('TupleToolDecayTreeFitter/ConsD')
dtt.Dstar.ConsD.constrainToOriginVertex = True
dtt.Dstar.ConsD.Verbose = True
dtt.Dstar.ConsD.daughtersToConstrain = ['D0']

# add another fitter, this time we will change a mass hypothesis
dtt.Dstar.addTupleTool('TupleToolDecayTreeFitter/ConsDKpi')
dtt.Dstar.ConsDKpi.constrainToOriginVertex = True
dtt.Dstar.ConsDKpi.Verbose = True
dtt.Dstar.ConsDKpi.daughtersToConstrain = ['D0']

# make the hypothesis that actually we had the decay D0->pi+pi-
# note that you have to explicitly give both charges
# CC does not work here!
dtt.Dstar.ConsDKpi.Substitutions = {
    'Charm -> (D0 -> ^K- K+) Meson': 'pi-',
    'Charm -> (D~0 -> ^K+ K-) Meson': 'pi+',
    'Charm -> (D0 -> K- ^K+) Meson': 'pi+',
    'Charm -> (D~0 -> K+ ^K-) Meson': 'pi-'
}


# Configure DaVinci
DaVinci().UserAlgorithms += [dtt]
DaVinci().PrintFreq = 1



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from subprocess import check_output

import pytest
from click.testing import CliRunner

from conftest import B02DKPi_versions
from LbAPLocal.cli import legacy_main as main

args_list = ["list"]
args_render = ["render", "cbtesting"]
args_test = ["test", "cbtesting", "MC_2011_MagUp_Lb2Lee_strip"]
args_debug = ["debug", "cbtesting", "MC_2011_MagUp_Lb2Lee_strip"]
args_reproduce = ["reproduce"]
args_parse_log = ["parse-log"]
args_versions = [["versions"], ["versions", "B2OC"], ["versions", "B2OC", "B02DKPi"]]
versions_output = [
    [
        "version querying currently only works with working group,"
        " and production name specified! lb-ap versions WORKING_GROUP PRODUCTION_NAME"
    ],
    [
        "version querying currently only works with working group,"
        " and production name specified! lb-ap versions WORKING_GROUP PRODUCTION_NAME"
    ],
    B02DKPi_versions,
]


args_clone_https = [
    "clone",
    "B2OC",
    "B02DKPi",
    "v0r0p2180615",
    "aiwieder/test_checkout",
    "--clone-type=https",
]
args_checkout = ["checkout"]


@pytest.mark.parametrize(
    "subcommand",
    [
        args_list,
        args_render,
        args_test,
        args_debug,
        args_reproduce,
        args_parse_log,
        *args_versions,
        args_clone_https,
        args_checkout,
    ],
)
def test_main_help(subcommand):
    runner = CliRunner()
    result = runner.invoke(main, subcommand + ["--help"])
    assert result.exit_code == 0, result.stdout


def test_main_version():
    runner = CliRunner()
    result = runner.invoke(main, ["--version"])
    assert result.exit_code == 0, result.stdout
    assert "LbAPLocal" in result.stdout
    assert "LbAPCommon" in result.stdout
    assert "LbEnv" in result.stdout
    assert "LbDiracWrappers" in result.stdout


@pytest.mark.parametrize("subcommand", [args_test, args_debug])
def test_not_in_datapkg(subcommand, empty_data_pkg_repo, with_proxy):
    runner = CliRunner()
    result = runner.invoke(main, subcommand)
    assert result.exit_code == 1
    assert "Running command in wrong directory" in result.output


@pytest.mark.parametrize(
    "subcommand",
    [
        args_test,
        args_debug,
    ],
)
def test_no_proxy(subcommand, data_pkg_repo, without_proxy):
    runner = CliRunner()
    result = runner.invoke(main, subcommand)
    assert result.exit_code == 1
    assert "No grid proxy found" in result.output
    assert "lhcb-proxy-init" in result.output


@pytest.mark.parametrize("subcommand", args_versions)
def test_versions(subcommand):
    runner = CliRunner()
    result = runner.invoke(main, subcommand)
    for output in versions_output[args_versions.index(subcommand)]:
        assert output in result.output


@pytest.mark.slow
@pytest.mark.parametrize("subcommand", [args_clone_https])
def test_clone_https(subcommand, tmpdir, monkeypatch, fake_global_git_config):
    monkeypatch.chdir(tmpdir)

    runner = CliRunner()
    result = runner.invoke(main, subcommand)
    assert result.exit_code == 0, result.output

    branch = check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
    assert branch.decode().strip() == "aiwieder/test_checkout"

    assert "Cloning into 'AnalysisProductions'..." in result.stdout
    assert "Checking out the v0r0p2180615 version of B02DKPi" in result.stdout




from cbtesting.helpers import tuple_maker

# ONLY CONFIG NEEDED
tuple_name     = 'Lb2JpsiL_mmTuple' # BECAUSE THIS RECONSTRUCTS Lb2Lee with Jpsi resonance

tuple_seq = tuple_maker.tuple_maker( tuple_name, upstream_electrons=False )




###############################################################################
# (c) Copyright 2026 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Textual-based TUI for CWL workflow execution monitoring.

Provides a real-time terminal UI for monitoring CWL workflow execution with:
- Workflow status and progress
- Live stdout from CWL executor
- Current step log file viewer
- Workflow graph visualization
- Output file tracking
- Memory usage monitoring
"""

import asyncio
import re
import time
from pathlib import Path
from typing import Optional

import yaml
from textual.app import App, ComposeResult
from textual.containers import Container, VerticalScroll
from textual.reactive import reactive
from textual.widgets import (
    DataTable,
    Footer,
    Header,
    RichLog,
    Static,
    TabbedContent,
    TabPane,
    Tree,
)
from textual_plotext import PlotextPlot


class WorkflowStatus(Static):
    """Header showing workflow execution status"""

    status = reactive("Starting...")
    elapsed_time = reactive("0s")
    current_step = reactive("Initializing...")
    memory_usage = reactive("")
    output_files_info = reactive("")
    max_memory = reactive("")

    def render(self) -> str:
        status_emoji = "🔄" if self.status == "Running" else "⏳"

        content = f"{status_emoji} [bold cyan]{self.status}[/bold cyan]"
        content += f"  │  ⏱️  {self.elapsed_time}"
        content += f"  │  📋 {self.current_step}"

        if self.output_files_info:
            content += f"  │  📂 {self.output_files_info}"

        if self.memory_usage:
            content += f"  │  💾 {self.memory_usage}"

        if self.max_memory:
            content += f"  │  📊 {self.max_memory}"

        return content


class TabbedLogViewer(Container):
    """Tabbed log file viewer for multiple steps"""

    DEFAULT_CSS = """
    TabbedLogViewer {
        border: solid $accent;
        height: 1fr;
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Logs"
        self.log_tabs = {}  # {log_file_path: state_dict}
        self.tabbed_content = None

    def compose(self) -> ComposeResult:
        """Build the tabbed interface"""
        yield TabbedContent()

    def on_mount(self):
        """Get reference to TabbedContent after mounting"""
        self.tabbed_content = self.query_one(TabbedContent)

    async def add_log_file(self, log_file: Path, step_name: str):
        """Add a new log file tab"""
        if not log_file or log_file in self.log_tabs:
            return

        if not self.tabbed_content:
            return

        # Create a short tab label
        tab_label = step_name[-20:] if len(step_name) > 20 else step_name
        tab_id = f"log_tab_{len(self.log_tabs) + 1}"

        # Create a simple TabPane
        pane = TabPane(tab_label, id=tab_id)
        await self.tabbed_content.add_pane(pane)

        # Mount a RichLog widget - designed for scrollable log content
        log_widget = RichLog(max_lines=500, auto_scroll=True, wrap=True)
        await pane.mount(log_widget)

        # Store reference with file tracking state
        self.log_tabs[log_file] = {
            "tab_id": tab_id,
            "widget": log_widget,
            "position": 0,
        }

        # Switch to the new tab
        self.tabbed_content.active = tab_id

    def refresh_all_logs(self):
        """Refresh all log file contents"""
        for log_file, state in self.log_tabs.items():
            if not log_file.exists():
                continue

            try:
                with open(log_file, "r") as f:
                    f.seek(state["position"])
                    new_lines = f.readlines()
                    if new_lines:
                        for line in new_lines:
                            state["widget"].write(line.rstrip())
                        state["position"] = f.tell()
            except Exception:
                pass


class MemoryChart(PlotextPlot):
    """Memory consumption chart using plotext for proper time series visualization"""

    DEFAULT_CSS = """
    MemoryChart {
        border: solid $primary;
        height: 100;
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Memory Usage (RSS+Swap)"
        self._times = []  # Time in seconds from start
        self._rss_values = []  # RSS in GB
        self._swap_values = []  # Swap in GB
        self._total_values = []  # RSS+Swap in GB
        self.max_memory = 0

    def on_mount(self):
        """Initialize the plot"""
        self._replot()

    def add_data_point(self, time_seconds: float, rss_kb: int, swap_kb: int = 0):
        """Add a new memory data point

        Args:
            time_seconds: Time in seconds from workflow start (relative time)
            rss_kb: RSS memory in KB
            swap_kb: Swap memory in KB
        """
        # Convert to GB
        rss_gb = rss_kb / (1024 * 1024)
        swap_gb = swap_kb / (1024 * 1024)
        total_gb = rss_gb + swap_gb

        # Store the data - time_seconds is already relative
        self._times.append(time_seconds)
        self._rss_values.append(rss_gb)
        self._swap_values.append(swap_gb)
        self._total_values.append(total_gb)

        if total_gb > self.max_memory:
            self.max_memory = total_gb

        # Limit data points to prevent memory issues
        max_points = 200
        if len(self._times) > max_points:
            self._times = self._times[-max_points:]
            self._rss_values = self._rss_values[-max_points:]
            self._swap_values = self._swap_values[-max_points:]
            self._total_values = self._total_values[-max_points:]

        self._replot()

    def _replot(self):
        """Redraw the plot with current data"""
        self.plt.clear_figure()

        if not self._times:
            self.plt.title("Waiting for memory data...")
            self.refresh()
            return

        # Plot RSS and Swap as stacked area effect using lines
        self.plt.plot(self._times, self._rss_values, label="RSS", marker="braille")
        if any(v > 0 for v in self._swap_values):
            self.plt.plot(
                self._times, self._total_values, label="RSS+Swap", marker="braille"
            )

        self.plt.xlabel("Time (s)")
        self.plt.ylabel("Memory (GB)")
        # self.plt.title("Memory Usage")

        self.refresh()


class WorkflowGraph(VerticalScroll):
    """Workflow structure visualization"""

    DEFAULT_CSS = """
    WorkflowGraph {
        border: solid $success;
        padding: 1;
    }
    """

    def __init__(self, workflow_steps: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Workflow Steps"
        self.workflow_steps = workflow_steps
        self._completed_steps = set()
        self._current_step_id = None
        self._tree = None
        self._step_outputs = {}  # Track files per step: {step_id: [(file, size), ...]}
        self._last_outputs_hash = None  # Track hash to detect changes

    def compose(self) -> ComposeResult:
        """Build the workflow graph tree"""
        self._tree = Tree("Workflow", id="workflow-tree")
        self._tree.root.expand()
        self._rebuild_tree()
        yield self._tree

    def _rebuild_tree(self):
        """Rebuild the tree with current status"""
        if not self._tree:
            return

        # Clear existing tree
        self._tree.clear()
        self._tree.root.expand()

        def add_steps_to_tree(parent_node, steps):
            for step in steps:
                step_id = step["id"]
                step_label = step.get("label", step_id)
                full_id = step.get("full_id", step_id)

                # Add status indicator
                if step_id in self._completed_steps:
                    icon = "✅"
                elif step_id == self._current_step_id:
                    icon = "▶️"
                else:
                    icon = "⏸️"

                node = parent_node.add(f"{icon} {step_label}", data=step)
                node.expand()

                # Add output files for this step
                step_files = []
                # Try multiple matching strategies (same as old Rich UI)
                for key in [step_id, full_id]:
                    if key in self._step_outputs:
                        step_files.extend(self._step_outputs[key])

                # Also try matching keys that contain step_id
                for key in self._step_outputs.keys():
                    if step_id in key or key in step_id:
                        if key not in [step_id, full_id]:
                            step_files.extend(self._step_outputs[key])

                # Add file nodes (deduplicate)
                seen_files = set()
                for file_path, size in step_files:
                    if file_path not in seen_files:
                        seen_files.add(file_path)
                        # Format size
                        if size > 1024 * 1024:
                            size_str = f"{size / (1024 * 1024):.1f} MB"
                        elif size > 1024:
                            size_str = f"{size / 1024:.1f} KB"
                        else:
                            size_str = f"{size} B"
                        node.add_leaf(f"📄 {file_path} ({size_str})")

                # Recursively add substeps
                if step.get("substeps"):
                    add_steps_to_tree(node, step["substeps"])

        add_steps_to_tree(self._tree.root, self.workflow_steps)

    def mark_step_running(self, step_id: str):
        """Mark a step as currently running"""
        self._current_step_id = step_id
        self._rebuild_tree()

    def mark_step_complete(self, step_id: str):
        """Mark a step as completed"""
        self._completed_steps.add(step_id)
        self._rebuild_tree()

    def update_step_outputs(self, step_outputs: dict):
        """Update the output files for all steps"""
        # Only rebuild if outputs actually changed
        # Use a fast hash-based comparison instead of JSON serialization
        try:
            # Create a simple hash from the dict structure
            outputs_repr = str(
                sorted((k, tuple(sorted(v))) for k, v in step_outputs.items())
            )
            outputs_hash = hash(outputs_repr)

            if outputs_hash != self._last_outputs_hash:
                self._last_outputs_hash = outputs_hash
                self._step_outputs = step_outputs.copy()
                self._rebuild_tree()
        except (TypeError, ValueError):
            # If hashing fails, always rebuild (safe fallback)
            self._step_outputs = step_outputs.copy()
            self._rebuild_tree()


class OutputFiles(DataTable):
    """Table showing output files and their sizes"""

    DEFAULT_CSS = """
    OutputFiles {
        border: solid $warning;
        height: 12;
    }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.border_title = "Output Files"
        self.files = {}

    def on_mount(self):
        """Set up the data table columns"""
        self.add_column("File", key="file")
        self.add_column("Size", key="size", width=15)
        self.add_column("Step", key="step", width=20)

    def update_files(self, files_dict: dict, step_outputs: dict):
        """Update the files table"""
        # Clear existing rows
        self.clear()

        # Add rows for each file
        for file_path, size in sorted(files_dict.items()):
            size_str = self._format_size(size)

            # Find which step produced this file
            step_name = "Unknown"
            for step, file_list in step_outputs.items():
                if any(fp == file_path for fp, _ in file_list):
                    step_name = step
                    break

            self.add_row(file_path, size_str, step_name)

    @staticmethod
    def _format_size(size_bytes: int) -> str:
        """Format file size in human-readable format"""
        for unit in ["B", "KB", "MB", "GB"]:
            if size_bytes < 1024:
                return f"{size_bytes:.1f} {unit}"
            size_bytes /= 1024
        return f"{size_bytes:.1f} TB"


class CWLMonitorApp(App):
    """Textual app for monitoring CWL workflow execution"""

    CSS = """
    Screen {
        layout: grid;
        grid-size: 2 4;
        grid-rows: auto 10 1fr 1fr;
        grid-columns: 30 1fr;
        height: 100vh;
    }

    WorkflowStatus {
        column-span: 2;
        height: 3;
        padding: 1;
        background: $boost;
    }

    MemoryChart {
        column-span: 2;
    }

    WorkflowGraph {
        row-span: 2;
    }

    TabbedLogViewer {
        row-span: 2;
    }

    OutputFiles {
        column-span: 2;
        height: auto;
    }
    """

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("s", "screenshot", "Screenshot"),
    ]

    def __init__(self, workflow_steps: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.workflow_steps = workflow_steps
        self.status_widget = None
        self.log_viewer = None
        self.graph_widget = None
        self.memory_chart = None
        self.output_files_widget = None
        self.start_time = None
        self.update_timer = None
        self.max_memory_kb = 0
        self.memory_warning_shown = False
        self.memory_critical_shown = False
        self.workflow_start_time = None
        self.workflow_completed = False

    def compose(self) -> ComposeResult:
        """Build the UI layout"""
        yield Header()

        self.status_widget = WorkflowStatus()
        yield self.status_widget

        self.memory_chart = MemoryChart()
        yield self.memory_chart

        self.graph_widget = WorkflowGraph(self.workflow_steps)
        yield self.graph_widget

        # Create log viewer (tabs will be added dynamically)
        self.log_viewer = TabbedLogViewer()
        yield self.log_viewer

        self.output_files_widget = OutputFiles()
        yield self.output_files_widget

        yield Footer()

    def on_mount(self):
        """Start the update timer when app mounts"""
        self.start_time = time.time()
        self.workflow_start_time = time.time()
        self.update_timer = self.set_interval(0.1, self.update_elapsed_time)
        self.log_refresh_timer = self.set_interval(0.2, self.refresh_log_viewer)
        # Force initial refresh to ensure UI is interactive immediately
        self.refresh()

    def update_elapsed_time(self):
        """Update the elapsed time display"""
        try:
            if self.start_time and self.status_widget and not self.workflow_completed:
                elapsed = time.time() - self.start_time
                mins = int(elapsed // 60)
                secs = int(elapsed % 60)
                self.status_widget.elapsed_time = f"{mins}m {secs}s"
        except Exception:
            # Silently ignore errors in timer updates
            pass

    def refresh_log_viewer(self):
        """Refresh all log viewer tabs"""
        try:
            if self.log_viewer:
                self.log_viewer.refresh_all_logs()
        except Exception:
            # Silently ignore errors in log refresh
            pass

    def set_current_step(self, step_name: str, step_id: Optional[str] = None):
        """Update the current step"""
        if self.status_widget:
            self.status_widget.current_step = step_name

        if step_id and self.graph_widget:
            self.graph_widget.mark_step_running(step_id)

    def mark_step_complete(self, step_id: str):
        """Mark a step as completed"""
        if self.graph_widget:
            self.graph_widget.mark_step_complete(step_id)

    def set_log_file(self, log_file: Optional[Path], step_name: str = "Step"):
        """Add a new log file tab"""
        if self.log_viewer and log_file:
            self.call_later(self.log_viewer.add_log_file, log_file, step_name)

    def update_files(self, files_dict: dict, step_outputs: dict):
        """Update the output files table and status info"""
        if self.output_files_widget:
            self.output_files_widget.update_files(files_dict, step_outputs)

        # Update workflow graph with step outputs
        if self.graph_widget:
            self.graph_widget.update_step_outputs(step_outputs)

        # Update output files info in status
        if self.status_widget and files_dict:
            file_count = len(files_dict)
            total_size = sum(files_dict.values())
            if total_size > 1024 * 1024:
                size_str = f"{total_size / (1024 * 1024):.1f} MB"
            elif total_size > 1024:
                size_str = f"{total_size / 1024:.1f} KB"
            else:
                size_str = f"{total_size} bytes"
            self.status_widget.output_files_info = f"{file_count} file(s) ({size_str})"

    def set_memory_usage(self, memory_str: str, memory_kb: int):
        """Update memory usage display and check thresholds"""
        if self.status_widget:
            self.status_widget.memory_usage = memory_str

            # Track max memory and update display
            if memory_kb > self.max_memory_kb:
                self.max_memory_kb = memory_kb
                max_memory_gb = memory_kb / (1024 * 1024)

                # Determine style and warnings
                if max_memory_gb >= 5.0:
                    warning_icon = " ⚠️  CRITICAL"
                    if not self.memory_critical_shown:
                        self.memory_critical_shown = True
                elif max_memory_gb >= 3.5:
                    warning_icon = " ⚠️"
                    if not self.memory_warning_shown:
                        self.memory_warning_shown = True
                else:
                    warning_icon = ""

                self.status_widget.max_memory = (
                    f"Max: {max_memory_gb:.1f} GB{warning_icon}"
                )

    def set_status(self, status: str):
        """Update workflow status"""
        if self.status_widget:
            self.status_widget.status = status

    def mark_workflow_complete(self, success: bool = True):
        """Mark workflow as completed to stop timers"""
        self.workflow_completed = True
        # Update final elapsed time
        if self.start_time and self.status_widget:
            elapsed = time.time() - self.start_time
            mins = int(elapsed // 60)
            secs = int(elapsed % 60)
            self.status_widget.elapsed_time = f"{mins}m {secs}s"

        # Show completion notification
        if success:
            self.notify(
                "✅ CWL workflow executed successfully!",
                title="Workflow Complete",
                severity="information",
                timeout=10,
            )
        else:
            self.notify(
                "❌ CWL workflow execution failed!",
                title="Workflow Failed",
                severity="error",
                timeout=10,
            )


async def run_cwl_with_textual_ui(cmd, env, output_dir, cwl_file):
    """
    Run CWL workflow with Textual TUI for monitoring.

    Args:
        cmd: Command list to execute
        env: Environment dictionary to pass to subprocess
        output_dir: CWL output directory where workflow output files are written
        cwl_file: Path to the CWL workflow file for graph visualization

    Returns:
        tuple: (exit_code, list of temp directories used, max_memory_kb)
    """

    # Parse CWL workflow to extract steps
    workflow_steps = []

    def parse_workflow_steps(workflow_data, parent_id=""):
        """Recursively parse workflow steps"""
        steps = []
        if "steps" not in workflow_data:
            return steps

        for step in workflow_data["steps"]:
            step_id = step.get("id", "unknown")
            full_id = f"{parent_id}/{step_id}" if parent_id else step_id

            run_section = step.get("run", {})

            if isinstance(run_section, dict):
                step_label = run_section.get("label", step_id)
                step_class = run_section.get("class", "unknown")

                steps.append(
                    {
                        "id": step_id,
                        "full_id": full_id,
                        "label": step_label,
                        "class": step_class,
                        "level": parent_id.count("/"),
                        "substeps": [],
                    }
                )

                if step_class == "Workflow":
                    substeps = parse_workflow_steps(run_section, full_id)
                    steps[-1]["substeps"] = substeps
            else:
                steps.append(
                    {
                        "id": step_id,
                        "full_id": full_id,
                        "label": step_id,
                        "class": "external",
                        "level": parent_id.count("/"),
                        "substeps": [],
                    }
                )

        return steps

    try:
        with open(cwl_file, "r") as f:
            cwl_data = yaml.safe_load(f)
            workflow_steps = parse_workflow_steps(cwl_data)
    except Exception:
        workflow_steps = [
            {
                "id": "workflow",
                "full_id": "workflow",
                "label": "Workflow",
                "class": "unknown",
                "level": 0,
                "substeps": [],
            }
        ]

    # Create the Textual app
    app = CWLMonitorApp(workflow_steps)

    # State tracking
    current_step_name = "Initializing..."
    current_log_file = None
    current_tmpdir = None
    current_output_prefix = None
    expecting_output_prefix = False
    temp_directories = []
    output_files = {}
    step_outputs = {}
    max_memory_usage = 0
    last_file_scan = 0
    file_scan_interval = 0.5
    process = None
    exit_code = 0
    prmon_data_points_read = {}  # Track lines read per prmon file to avoid duplicates

    # Patterns for log parsing
    step_pattern = re.compile(r"\[(job|step|workflow) ([^\]]+)\]")
    job_pattern = re.compile(r"\[job\s+([^\]]+)\]\s+(/tmp/[^\s$]+)\$")
    workflow_step_pattern = re.compile(
        r"\[(workflow|step)\s+([^\]]+)\]\s+(starting step|start)"
    )
    # tmpdir_pattern = re.compile(r"tmp[a-z0-9_-]{6,}", re.IGNORECASE)
    # output_prefix_pattern = re.compile(r"--output-prefix")
    # prmon_pattern = re.compile(r"prmon.*\.txt")

    def find_current_log_file(line):
        """Extract tmpdir and output_prefix from CWL executor output to locate log file"""
        nonlocal current_log_file, current_step_name, current_tmpdir
        nonlocal current_output_prefix, expecting_output_prefix, temp_directories

        # Pattern 1a: Look for workflow/step starting messages
        workflow_step_match = workflow_step_pattern.search(line)
        if workflow_step_match:
            step = workflow_step_match.group(2)
            current_step_name = step

        # Pattern 1b: Look for cwltool job execution messages
        job_match = job_pattern.search(line)
        if job_match:
            step = job_match.group(1)
            tmpdir = Path(job_match.group(2))

            # Update tracking variables
            current_step_name = step
            current_tmpdir = tmpdir
            # Track this temp directory for later log file discovery
            if tmpdir not in temp_directories:
                temp_directories.append(tmpdir)

        # Pattern 2: Look for --output-prefix argument
        if "--output-prefix" in line:
            # Next non-empty, non-backslash line will have the prefix value
            expecting_output_prefix = True
        elif expecting_output_prefix:
            # This line should contain the actual output prefix value
            stripped = line.strip()
            # Skip lines that are just backslashes or empty
            if stripped and stripped != "\\":
                current_output_prefix = stripped
                expecting_output_prefix = False

        # Pattern 3: If we have both tmpdir and output_prefix, try to find the log file
        if current_tmpdir and current_output_prefix and current_tmpdir.exists():
            # The log file pattern is: {tmpdir}/*_{output_prefix}.log
            log_pattern = f"*_{current_output_prefix}.log"
            log_files = list(current_tmpdir.glob(log_pattern))

            if log_files:
                # Use the first (should be only one) matching log file
                if log_files[0] != current_log_file:
                    current_log_file = log_files[0]
                    app.set_log_file(current_log_file, current_step_name)
                    return True

        return False

    async def monitor_process():
        """Monitor the CWL process and update UI"""
        nonlocal process, exit_code, current_step_name, current_log_file, current_tmpdir
        nonlocal temp_directories, output_files, step_outputs, max_memory_usage, last_file_scan
        nonlocal current_output_prefix, expecting_output_prefix, prmon_data_points_read

        # Wait for app to be ready (widgets mounted)
        while not app.log_viewer:
            await asyncio.sleep(0.1)

        # Create executor log file in output directory
        executor_log_path = output_dir / "executor.log"
        executor_log_file = open(executor_log_path, "w", buffering=1)  # Line buffered

        # Add executor log to the UI (now that app is ready)
        app.set_log_file(executor_log_path, "CWL Executor")

        # Start the subprocess
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            env=env,
        )

        app.set_status("Running")

        # Read stdout line by line
        async for line_bytes in process.stdout:
            line = line_bytes.decode("utf-8", errors="replace").rstrip()

            # Write to executor log file
            executor_log_file.write(line + "\n")
            executor_log_file.flush()

            # Try to find log file references in the output
            find_current_log_file(line)

            # Parse for step information to update the graph
            step_match = step_pattern.search(line)
            if step_match:
                step_type = step_match.group(1)
                step_id = step_match.group(2)
                app.set_current_step(f"[{step_type} {step_id}]", step_id)

            # Periodic file scanning
            current_time = time.time()
            if current_time - last_file_scan >= file_scan_interval:
                last_file_scan = current_time

                # Scan for output files
                new_files = {}
                current_step_files = []

                # Scan final output directory
                for ext in ["*.root", "*.dst", "*.xdigi", "*.xml", "*.json", "*.txt"]:
                    for f in output_dir.glob(f"**/{ext}"):
                        if f.is_file():
                            size = f.stat().st_size
                            rel_path = str(f.relative_to(output_dir))
                            new_files[rel_path] = size

                            # Track new files for current step
                            if (
                                rel_path not in output_files
                                and current_step_name != "Initializing..."
                            ):
                                current_step_files.append((rel_path, size))

                # Also scan tmpdir
                if current_tmpdir and current_tmpdir.exists():
                    for ext in [
                        "*.root",
                        "*.dst",
                        "*.xdigi",
                        "*.xml",
                        "*.json",
                        "*.txt",
                    ]:
                        for f in current_tmpdir.glob(ext):
                            if f.is_file():
                                size = f.stat().st_size
                                file_key = f.name  # Use just the filename as key
                                new_files[file_key] = size

                                # Track new files for the current step
                                if (
                                    file_key not in output_files
                                    and current_step_name != "Initializing..."
                                ):
                                    current_step_files.append((file_key, size))

                # Add new files to the current step's output list
                if current_step_files:
                    clean_step_name = (
                        current_step_name.replace("[job ", "")
                        .replace("[step ", "")
                        .replace("[workflow ", "")
                        .replace("]", "")
                        .strip()
                    )
                    if clean_step_name not in step_outputs:
                        step_outputs[clean_step_name] = []
                    step_outputs[clean_step_name].extend(current_step_files)

                # Update sizes of existing files in step_outputs
                for _, files_list in step_outputs.items():
                    for i, (file_path, _) in enumerate(files_list):
                        if file_path in new_files:
                            # Update the size tuple in place
                            files_list[i] = (file_path, new_files[file_path])

                output_files = new_files
                app.update_files(output_files, step_outputs)

                # Check memory usage from prmon files
                for tmpdir in temp_directories:
                    if tmpdir.exists():
                        prmon_files = list(tmpdir.glob("prmon*.txt"))
                        for prmon_file in prmon_files:
                            try:
                                content = prmon_file.read_text()
                                # Parse all lines for time-series data (Time, RSS, PSS, Swap, ...)
                                lines = [
                                    line
                                    for line in content.splitlines()
                                    if line and not line.startswith("#")
                                ]

                                if not lines:
                                    continue

                                # Track which lines we've already processed for this file
                                prmon_key = str(prmon_file)
                                lines_already_read = prmon_data_points_read.get(
                                    prmon_key, 0
                                )

                                # Process only new lines to avoid duplicates
                                current_max_mem = 0
                                for i, line in enumerate(lines):
                                    # Skip lines we've already processed
                                    if i < lines_already_read:
                                        continue

                                    parts = line.split()
                                    if (
                                        len(parts) >= 4
                                    ):  # Time,1 WTime, 2 RSS, 3 PSS, 4 Swap, ...
                                        try:
                                            timestamp_seconds = float(
                                                parts[1]
                                            )  # Time in seconds from start
                                            rss = float(parts[3])  # KB
                                            swap = (
                                                float(parts[4]) if len(parts) > 3 else 0
                                            )  # KB
                                            total_mem = rss + swap

                                            # Add data point to chart - timestamp_seconds is already relative
                                            if app.memory_chart:
                                                app.memory_chart.add_data_point(
                                                    timestamp_seconds,
                                                    int(rss),
                                                    int(swap),
                                                )

                                            # Track max for this scan
                                            if total_mem > current_max_mem:
                                                current_max_mem = total_mem
                                        except (ValueError, IndexError):
                                            # Skip malformed lines
                                            continue

                                # Update how many lines we've read for this file
                                prmon_data_points_read[prmon_key] = len(lines)

                                # Update status widget with latest max
                                if current_max_mem > max_memory_usage:
                                    max_memory_usage = current_max_mem
                                    mem_gb = current_max_mem / (1024 * 1024)
                                    mem_str = f"RSS: {mem_gb:.2f} GB"
                                    app.set_memory_usage(mem_str, int(current_max_mem))
                            except Exception:
                                # Silently ignore prmon read errors
                                pass

        # Wait for process to complete
        exit_code = await process.wait()

        # Close executor log file
        executor_log_file.close()

        if exit_code == 0:
            app.set_status("✅ Completed")
            # Mark workflow as complete with success
            app.mark_workflow_complete(success=True)
        else:
            app.set_status(f"❌ Failed (exit {exit_code})")
            # Mark workflow as complete with failure
            app.mark_workflow_complete(success=False)

    # Run the app and process monitoring concurrently
    async def run_app():
        await app.run_async()

    # Start both tasks
    monitor_task = asyncio.create_task(monitor_process())
    app_task = asyncio.create_task(run_app())

    # Wait for both to complete, catching any exceptions
    try:
        await asyncio.gather(monitor_task, app_task, return_exceptions=False)
    except Exception as e:
        # If an exception occurs, ensure we still return properly
        import sys

        print(f"Error in workflow execution: {e}", file=sys.stderr)

    return exit_code, temp_directories, int(max_memory_usage)
